from eletter import __version__

project = "eletter"
author = "John Thorvald Wodder II"
copyright = "2021 John Thorvald Wodder II"

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx_copybutton",
]

autodoc_default_options = {
    "members": True,
    "undoc-members": True,
}

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
}

exclude_patterns = ["_build"]
source_suffix = ".rst"
source_encoding = "utf-8"
master_doc = "index"
version = __version__
release = __version__
today_fmt = "%Y %b %d"
default_role = "py:obj"
pygments_style = "sphinx"
todo_include_todos = True

html_theme = "sphinx_rtd_theme"
html_theme_options = {
    "collapse_navigation": False,
    "prev_next_buttons_location": "both",
}
html_last_updated_fmt = "%Y %b %d"
html_show_sourcelink = True
html_show_sphinx = True
html_show_copyright = True



from datetime import datetime, timedelta, timezone
from mailbits import email2dict
import pytest
from eletter import Address, Group, compose


def test_compose_text() -> None:
    msg = compose(
        from_="me@here.com",
        to=["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
        subject="Some electronic mail",
        text="This is the text of an e-mail.",
    )
    assert email2dict(msg) == {
        "unixfrom": None,
        "headers": {
            "subject": "Some electronic mail",
            "from": [
                {
                    "display_name": "",
                    "address": "me@here.com",
                }
            ],
            "to": [
                {
                    "display_name": "",
                    "address": "you@there.net",
                },
                {
                    "display_name": "Thaddeus Hem",
                    "address": "them@hither.yon",
                },
            ],
            "content-type": {
                "content_type": "text/plain",
                "params": {},
            },
        },
        "preamble": None,
        "content": "This is the text of an e-mail.\n",
        "epilogue": None,
    }


def test_compose_html() -> None:
    msg = compose(
        from_="me@here.com",
        to=["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
        subject="Some electronic mail",
        html="<p>This is the <i>text</i> of an <b>e</b>-mail.<p>",
    )
    assert email2dict(msg) == {
        "unixfrom": None,
        "headers": {
            "subject": "Some electronic mail",
            "from": [
                {
                    "display_name": "",
                    "address": "me@here.com",
                }
            ],
            "to": [
                {
                    "display_name": "",
                    "address": "you@there.net",
                },
                {
                    "display_name": "Thaddeus Hem",
                    "address": "them@hither.yon",
                },
            ],
            "content-type": {
                "content_type": "text/html",
                "params": {},
            },
        },
        "preamble": None,
        "content": "<p>This is the <i>text</i> of an <b>e</b>-mail.<p>\n",
        "epilogue": None,
    }


def test_compose_text_html() -> None:
    msg = compose(
        from_="me@here.com",
        to=["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
        subject="Some electronic mail",
        text="This is the text of an e-mail.",
        html="<p>This is the <i>text</i> of an <b>e</b>-mail.<p>",
    )
    assert email2dict(msg) == {
        "unixfrom": None,
        "headers": {
            "subject": "Some electronic mail",
            "from": [
                {
                    "display_name": "",
                    "address": "me@here.com",
                }
            ],
            "to": [
                {
                    "display_name": "",
                    "address": "you@there.net",
                },
                {
                    "display_name": "Thaddeus Hem",
                    "address": "them@hither.yon",
                },
            ],
            "content-type": {
                "content_type": "multipart/alternative",
                "params": {},
            },
        },
        "preamble": None,
        "content": [
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "text/plain",
                        "params": {},
                    },
                },
                "preamble": None,
                "content": "This is the text of an e-mail.\n",
                "epilogue": None,
            },
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "text/html",
                        "params": {},
                    },
                },
                "preamble": None,
                "content": "<p>This is the <i>text</i> of an <b>e</b>-mail.<p>\n",
                "epilogue": None,
            },
        ],
        "epilogue": None,
    }


def test_compose_neither() -> None:
    with pytest.raises(ValueError) as excinfo:
        compose(
            from_="me@here.com",
            to=["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
            subject="Some electronic mail",
        )
    assert str(excinfo.value) == "At least one of text and html must be non-None"


def test_compose_addresses() -> None:
    msg = compose(
        from_=Address("Mme E.", "me@here.com"),
        to=["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
        cc=[Address("Cee Cee Cecil", "ccc@seesaws.cc"), "coco@nu.tz"],
        bcc=[
            "eletter@depository.nil",
            Address("Secret Cabal", "illuminati@new.world.order"),
            "mom@house.home",
        ],
        reply_to="replyee@some.where",
        sender="steven.ender@big.senders",
        subject="To: Everyone",
        text="This is the text of an e-mail.",
    )
    assert email2dict(msg) == {
        "unixfrom": None,
        "headers": {
            "subject": "To: Everyone",
            "from": [
                {
                    "display_name": "Mme E.",
                    "address": "me@here.com",
                }
            ],
            "to": [
                {
                    "display_name": "",
                    "address": "you@there.net",
                },
                {
                    "display_name": "Thaddeus Hem",
                    "address": "them@hither.yon",
                },
            ],
            "cc": [
                {
                    "display_name": "Cee Cee Cecil",
                    "address": "ccc@seesaws.cc",
                },
                {
                    "display_name": "",
                    "address": "coco@nu.tz",
                },
            ],
            "bcc": [
                {
                    "display_name": "",
                    "address": "eletter@depository.nil",
                },
                {
                    "display_name": "Secret Cabal",
                    "address": "illuminati@new.world.order",
                },
                {
                    "display_name": "",
                    "address": "mom@house.home",
                },
            ],
            "reply-to": [
                {
                    "display_name": "",
                    "address": "replyee@some.where",
                },
            ],
            "sender": {
                "display_name": "",
                "address": "steven.ender@big.senders",
            },
            "content-type": {
                "content_type": "text/plain",
                "params": {},
            },
        },
        "preamble": None,
        "content": "This is the text of an e-mail.\n",
        "epilogue": None,
    }


def test_compose_date() -> None:
    when = datetime(2021, 3, 8, 18, 14, 29, tzinfo=timezone(timedelta(hours=-5)))
    msg = compose(
        from_="me@here.com",
        to=["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
        subject="Some electronic mail",
        text="This is the text of an e-mail.",
        date=when,
    )
    assert email2dict(msg) == {
        "unixfrom": None,
        "headers": {
            "subject": "Some electronic mail",
            "from": [
                {
                    "display_name": "",
                    "address": "me@here.com",
                }
            ],
            "to": [
                {
                    "display_name": "",
                    "address": "you@there.net",
                },
                {
                    "display_name": "Thaddeus Hem",
                    "address": "them@hither.yon",
                },
            ],
            "date": when,
            "content-type": {
                "content_type": "text/plain",
                "params": {},
            },
        },
        "preamble": None,
        "content": "This is the text of an e-mail.\n",
        "epilogue": None,
    }


def test_compose_headers() -> None:
    msg = compose(
        from_="me@here.com",
        to=["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
        subject="Some electronic mail",
        text="This is the text of an e-mail.",
        headers={
            "User-Agent": "eletter",
            "Received": [
                "From mx.example.com",
                "From mail.sender.nil",
            ],
        },
    )
    assert email2dict(msg) == {
        "unixfrom": None,
        "headers": {
            "subject": "Some electronic mail",
            "from": [
                {
                    "display_name": "",
                    "address": "me@here.com",
                }
            ],
            "to": [
                {
                    "display_name": "",
                    "address": "you@there.net",
                },
                {
                    "display_name": "Thaddeus Hem",
                    "address": "them@hither.yon",
                },
            ],
            "user-agent": ["eletter"],
            "received": [
                "From mx.example.com",
                "From mail.sender.nil",
            ],
            "content-type": {
                "content_type": "text/plain",
                "params": {},
            },
        },
        "preamble": None,
        "content": "This is the text of an e-mail.\n",
        "epilogue": None,
    }


def test_compose_list_from_list_reply_to() -> None:
    msg = compose(
        from_=[Address("Mme E.", "me@here.com"), "also-me@hence.thither"],
        to=["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
        reply_to=[
            "replyee@some.where",
            Address("Response Handler", "r.handler@answers.4.you"),
        ],
        subject="Some electronic mail",
        text="This is the text of an e-mail.",
    )
    assert email2dict(msg) == {
        "unixfrom": None,
        "headers": {
            "subject": "Some electronic mail",
            "from": [
                {
                    "display_name": "Mme E.",
                    "address": "me@here.com",
                },
                {
                    "display_name": "",
                    "address": "also-me@hence.thither",
                },
            ],
            "to": [
                {
                    "display_name": "",
                    "address": "you@there.net",
                },
                {
                    "display_name": "Thaddeus Hem",
                    "address": "them@hither.yon",
                },
            ],
            "reply-to": [
                {
                    "display_name": "",
                    "address": "replyee@some.where",
                },
                {
                    "display_name": "Response Handler",
                    "address": "r.handler@answers.4.you",
                },
            ],
            "content-type": {
                "content_type": "text/plain",
                "params": {},
            },
        },
        "preamble": None,
        "content": "This is the text of an e-mail.\n",
        "epilogue": None,
    }


def test_compose_group() -> None:
    msg = compose(
        from_=Address("Mme E.", "me@here.com"),
        to=[
            Group(
                "friends", ["you@there.net", Address("Thaddeus Hem", "them@hither.yon")]
            ),
            "those@ovar.thar",
        ],
        subject="Some electronic mail",
        text="This is the text of an e-mail.",
    )
    assert email2dict(msg) == {
        "unixfrom": None,
        "headers": {
            "subject": "Some electronic mail",
            "from": [
                {
                    "display_name": "Mme E.",
                    "address": "me@here.com",
                },
            ],
            "to": [
                {
                    "group": "friends",
                    "addresses": [
                        {
                            "display_name": "",
                            "address": "you@there.net",
                        },
                        {
                            "display_name": "Thaddeus Hem",
                            "address": "them@hither.yon",
                        },
                    ],
                },
                {
                    "display_name": "",
                    "address": "those@ovar.thar",
                },
            ],
            "content-type": {
                "content_type": "text/plain",
                "params": {},
            },
        },
        "preamble": None,
        "content": "This is the text of an e-mail.\n",
        "epilogue": None,
    }


def test_compose_null_from() -> None:
    msg = compose(
        to=["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
        subject="Some electronic mail",
        text="This is the text of an e-mail.",
    )
    assert email2dict(msg) == {
        "unixfrom": None,
        "headers": {
            "subject": "Some electronic mail",
            "to": [
                {
                    "display_name": "",
                    "address": "you@there.net",
                },
                {
                    "display_name": "Thaddeus Hem",
                    "address": "them@hither.yon",
                },
            ],
            "content-type": {
                "content_type": "text/plain",
                "params": {},
            },
        },
        "preamble": None,
        "content": "This is the text of an e-mail.\n",
        "epilogue": None,
    }



from datetime import datetime, timedelta, timezone
from email.utils import make_msgid
from typing import Any, Iterable, Type, Union
from mailbits import email2dict
import pytest
from eletter import (
    Address,
    Alternative,
    BytesAttachment,
    Group,
    HTMLBody,
    MailItem,
    Mixed,
    Multipart,
    Related,
    TextAttachment,
    TextBody,
)
from eletter.util import AddressOrGroup


def test_text_body() -> None:
    msg = TextBody("This is the text of an e-mail.").compose(
        from_="me@here.com",
        to=["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
        subject="Some electronic mail",
    )
    assert email2dict(msg) == {
        "unixfrom": None,
        "headers": {
            "subject": "Some electronic mail",
            "from": [
                {
                    "display_name": "",
                    "address": "me@here.com",
                }
            ],
            "to": [
                {
                    "display_name": "",
                    "address": "you@there.net",
                },
                {
                    "display_name": "Thaddeus Hem",
                    "address": "them@hither.yon",
                },
            ],
            "content-type": {
                "content_type": "text/plain",
                "params": {},
            },
        },
        "preamble": None,
        "content": "This is the text of an e-mail.\n",
        "epilogue": None,
    }


def test_html_body() -> None:
    msg = HTMLBody("<p>This is the <i>text</i> of an <b>e</b>-mail.<p>").compose(
        from_="me@here.com",
        to=["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
        subject="Some electronic mail",
    )
    assert email2dict(msg) == {
        "unixfrom": None,
        "headers": {
            "subject": "Some electronic mail",
            "from": [
                {
                    "display_name": "",
                    "address": "me@here.com",
                }
            ],
            "to": [
                {
                    "display_name": "",
                    "address": "you@there.net",
                },
                {
                    "display_name": "Thaddeus Hem",
                    "address": "them@hither.yon",
                },
            ],
            "content-type": {
                "content_type": "text/html",
                "params": {},
            },
        },
        "preamble": None,
        "content": "<p>This is the <i>text</i> of an <b>e</b>-mail.<p>\n",
        "epilogue": None,
    }


def test_text_alt_html() -> None:
    t = TextBody("This is the text of an e-mail.")
    h = HTMLBody("<p>This is the <i>text</i> of an <b>e</b>-mail.<p>")
    msg = (t | h).compose(
        from_="me@here.com",
        to=["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
        subject="Some electronic mail",
    )
    assert email2dict(msg) == {
        "unixfrom": None,
        "headers": {
            "subject": "Some electronic mail",
            "from": [
                {
                    "display_name": "",
                    "address": "me@here.com",
                }
            ],
            "to": [
                {
                    "display_name": "",
                    "address": "you@there.net",
                },
                {
                    "display_name": "Thaddeus Hem",
                    "address": "them@hither.yon",
                },
            ],
            "content-type": {
                "content_type": "multipart/alternative",
                "params": {},
            },
        },
        "preamble": None,
        "content": [
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "text/plain",
                        "params": {},
                    },
                },
                "preamble": None,
                "content": "This is the text of an e-mail.\n",
                "epilogue": None,
            },
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "text/html",
                        "params": {},
                    },
                },
                "preamble": None,
                "content": "<p>This is the <i>text</i> of an <b>e</b>-mail.<p>\n",
                "epilogue": None,
            },
        ],
        "epilogue": None,
    }


def test_html_alt_text() -> None:
    h = HTMLBody("<p>This is the <i>text</i> of an <b>e</b>-mail.<p>")
    t = TextBody("This is the text of an e-mail.")
    msg = (h | t).compose(
        from_="me@here.com",
        to=["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
        subject="Some electronic mail",
    )
    assert email2dict(msg) == {
        "unixfrom": None,
        "headers": {
            "subject": "Some electronic mail",
            "from": [
                {
                    "display_name": "",
                    "address": "me@here.com",
                }
            ],
            "to": [
                {
                    "display_name": "",
                    "address": "you@there.net",
                },
                {
                    "display_name": "Thaddeus Hem",
                    "address": "them@hither.yon",
                },
            ],
            "content-type": {
                "content_type": "multipart/alternative",
                "params": {},
            },
        },
        "preamble": None,
        "content": [
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "text/html",
                        "params": {},
                    },
                },
                "preamble": None,
                "content": "<p>This is the <i>text</i> of an <b>e</b>-mail.<p>\n",
                "epilogue": None,
            },
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "text/plain",
                        "params": {},
                    },
                },
                "preamble": None,
                "content": "This is the text of an e-mail.\n",
                "epilogue": None,
            },
        ],
        "epilogue": None,
    }


def test_alt() -> None:
    t = TextBody("This is the text of an e-mail.")
    h = HTMLBody("<p>This is the <i>text</i> of an <b>e</b>-mail.<p>")
    alt = t | h
    assert isinstance(alt, Alternative)
    assert alt == Alternative([t, h])


def test_or_alt() -> None:
    t1 = TextBody("Part 1")
    t2 = TextBody("Part 2")
    t3 = TextBody("Part 3")
    t4 = TextBody("Part 4")
    alt1 = t1 | t2
    alt2 = t3 | t4
    alt = alt1 | alt2
    assert alt == Alternative([t1, t2, t3, t4])


def test_alt_or_eq_body() -> None:
    t1 = TextBody("Part 1")
    t2 = TextBody("Part 2")
    t3 = TextBody("Part 3")
    alt = t1 | t2
    x = alt
    x |= t3
    assert x is alt
    assert x == Alternative([t1, t2, t3])
    assert t3 == TextBody("Part 3")


def test_alt_or_eq_alt() -> None:
    t1 = TextBody("Part 1")
    t2 = TextBody("Part 2")
    t3 = TextBody("Part 3")
    t4 = TextBody("Part 4")
    alt1 = t1 | t2
    alt2 = t3 | t4
    x = alt1
    x |= alt2
    assert x is alt1
    assert x == Alternative([t1, t2, t3, t4])
    assert alt2 == Alternative([t3, t4])


def test_body_or_eq() -> None:
    t = TextBody("This is the text of an e-mail.")
    h = HTMLBody("<p>This is the <i>text</i> of an <b>e</b>-mail.<p>")
    x: Any = t
    x |= h
    assert x is not t
    assert isinstance(x, Alternative)
    assert x == Alternative([t, h])
    assert t == TextBody("This is the text of an e-mail.")


def test_compose_empty_alt() -> None:
    with pytest.raises(ValueError) as excinfo:
        Alternative([]).compose(
            from_="me@here.com",
            to=["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
            subject="Some electronic mail",
        )
    assert str(excinfo.value) == "Cannot compose empty Alternative"


@pytest.mark.parametrize(
    "mi",
    [
        TextBody("This is the text of an e-mail."),
        HTMLBody("<p>This is the <i>text</i> of an <b>e</b>-mail.<p>"),
        Alternative(
            [
                TextBody("This is the text of an e-mail."),
                HTMLBody("<p>This is the <i>text</i> of an <b>e</b>-mail.<p>"),
            ]
        ),
        Mixed(
            [
                TextBody("This is the text of an e-mail."),
                TextAttachment(
                    "this,text,attachment", "attachment.csv", content_type="text/csv"
                ),
            ]
        ),
        Related(
            [
                TextBody("This is the text of an e-mail."),
                TextAttachment(
                    "this,text,attachment", "attachment.csv", content_type="text/csv"
                ),
            ]
        ),
        TextAttachment("This is an attachment.", filename="foo.txt"),
        BytesAttachment(b"This is an attachment.\n", filename="foo.blob"),
    ],
)
@pytest.mark.parametrize(
    "from_input,from_output",
    [
        (None, None),
        (
            Address("Mme E.", "me@here.com"),
            [{"display_name": "Mme E.", "address": "me@here.com"}],
        ),
        (
            [Address("Mme E.", "me@here.com"), "also-me@hence.thither"],
            [
                {"display_name": "Mme E.", "address": "me@here.com"},
                {"display_name": "", "address": "also-me@hence.thither"},
            ],
        ),
    ],
)
@pytest.mark.parametrize(
    "to_input,to_output",
    [
        (
            ["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
            [
                {"display_name": "", "address": "you@there.net"},
                {"display_name": "Thaddeus Hem", "address": "them@hither.yon"},
            ],
        ),
        (
            [
                Group(
                    "friends",
                    ["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
                ),
                "those@ovar.thar",
            ],
            [
                {
                    "group": "friends",
                    "addresses": [
                        {"display_name": "", "address": "you@there.net"},
                        {"display_name": "Thaddeus Hem", "address": "them@hither.yon"},
                    ],
                },
                {"display_name": "", "address": "those@ovar.thar"},
            ],
        ),
    ],
)
@pytest.mark.parametrize(
    "reply_to_input,reply_to_output",
    [
        (
            "replyee@some.where",
            [{"display_name": "", "address": "replyee@some.where"}],
        ),
        (
            [
                "replyee@some.where",
                Address("Response Handler", "r.handler@answers.4.you"),
            ],
            [
                {"display_name": "", "address": "replyee@some.where"},
                {
                    "display_name": "Response Handler",
                    "address": "r.handler@answers.4.you",
                },
            ],
        ),
    ],
)
def test_mailitem_compose(
    mi: MailItem,
    from_input: Union[AddressOrGroup, Iterable[AddressOrGroup]],
    from_output: Any,
    to_input: Iterable[AddressOrGroup],
    to_output: Any,
    reply_to_input: Union[AddressOrGroup, Iterable[AddressOrGroup]],
    reply_to_output: Any,
) -> None:
    when = datetime(2021, 3, 8, 18, 14, 29, tzinfo=timezone(timedelta(hours=-5)))
    msg = mi.compose(
        subject="To: Everyone",
        from_=from_input,
        to=to_input,
        cc=[Address("Cee Cee Cecil", "ccc@seesaws.cc"), "coco@nu.tz"],
        bcc=[
            "eletter@depository.nil",
            Address("Secret Cabal", "illuminati@new.world.order"),
            "mom@house.home",
        ],
        reply_to=reply_to_input,
        sender="steven.ender@big.senders",
        date=when,
        headers={
            "User-Agent": "eletter",
            "Received": [
                "From mx.example.com",
                "From mail.sender.nil",
            ],
        },
    )
    headers = email2dict(msg)["headers"]
    headers.pop("content-type")
    headers.pop("content-disposition", None)
    expected = {
        "subject": "To: Everyone",
        "from": from_output,
        "to": to_output,
        "cc": [
            {
                "display_name": "Cee Cee Cecil",
                "address": "ccc@seesaws.cc",
            },
            {
                "display_name": "",
                "address": "coco@nu.tz",
            },
        ],
        "bcc": [
            {
                "display_name": "",
                "address": "eletter@depository.nil",
            },
            {
                "display_name": "Secret Cabal",
                "address": "illuminati@new.world.order",
            },
            {
                "display_name": "",
                "address": "mom@house.home",
            },
        ],
        "reply-to": reply_to_output,
        "sender": {
            "display_name": "",
            "address": "steven.ender@big.senders",
        },
        "date": when,
        "user-agent": ["eletter"],
        "received": [
            "From mx.example.com",
            "From mail.sender.nil",
        ],
    }
    if from_output is None:
        expected.pop("from")
    assert headers == expected


def test_text_and_attachment() -> None:
    t = TextBody("This is the text of an e-mail.")
    a = TextAttachment(
        "this,text,attachment", "attachment.csv", content_type="text/csv"
    )
    msg = (t & a).compose(
        from_="me@here.com",
        to=["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
        subject="Some electronic mail",
    )
    assert email2dict(msg) == {
        "unixfrom": None,
        "headers": {
            "subject": "Some electronic mail",
            "from": [
                {
                    "display_name": "",
                    "address": "me@here.com",
                }
            ],
            "to": [
                {
                    "display_name": "",
                    "address": "you@there.net",
                },
                {
                    "display_name": "Thaddeus Hem",
                    "address": "them@hither.yon",
                },
            ],
            "content-type": {
                "content_type": "multipart/mixed",
                "params": {},
            },
        },
        "preamble": None,
        "content": [
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "text/plain",
                        "params": {},
                    },
                },
                "preamble": None,
                "content": "This is the text of an e-mail.\n",
                "epilogue": None,
            },
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "text/csv",
                        "params": {},
                    },
                    "content-disposition": {
                        "disposition": "attachment",
                        "params": {"filename": "attachment.csv"},
                    },
                },
                "preamble": None,
                "content": "this,text,attachment\n",
                "epilogue": None,
            },
        ],
        "epilogue": None,
    }


def test_mixed() -> None:
    t = TextBody("This is the text of an e-mail.")
    a = TextAttachment(
        "this,text,attachment", "attachment.csv", content_type="text/csv"
    )
    mixed = t & a
    assert isinstance(mixed, Mixed)
    assert mixed == Mixed([t, a])


def test_and_mixed() -> None:
    t1 = TextBody("Part 1")
    t2 = TextBody("Part 2")
    t3 = TextBody("Part 3")
    t4 = TextBody("Part 4")
    mixed1 = t1 & t2
    mixed2 = t3 & t4
    mixed = mixed1 & mixed2
    assert mixed == Mixed([t1, t2, t3, t4])


def test_mixed_and_eq_body() -> None:
    t1 = TextBody("Part 1")
    t2 = TextBody("Part 2")
    t3 = TextBody("Part 3")
    mixed = t1 & t2
    x = mixed
    x &= t3
    assert x is mixed
    assert x == Mixed([t1, t2, t3])
    assert t3 == TextBody("Part 3")


def test_mixed_and_eq_mixed() -> None:
    t1 = TextBody("Part 1")
    t2 = TextBody("Part 2")
    t3 = TextBody("Part 3")
    t4 = TextBody("Part 4")
    mixed1 = t1 & t2
    mixed2 = t3 & t4
    x = mixed1
    x &= mixed2
    assert x is mixed1
    assert x == Mixed([t1, t2, t3, t4])
    assert mixed2 == Mixed([t3, t4])


def test_body_and_eq() -> None:
    t = TextBody("This is the text of an e-mail.")
    a = TextAttachment(
        "this,text,attachment", "attachment.csv", content_type="text/csv"
    )
    x: Any = t
    x &= a
    assert x is not t
    assert isinstance(x, Mixed)
    assert x == Mixed([t, a])
    assert t == TextBody("This is the text of an e-mail.")


def test_compose_empty_mixed() -> None:
    with pytest.raises(ValueError) as excinfo:
        Mixed([]).compose(
            from_="me@here.com",
            to=["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
            subject="Some electronic mail",
        )
    assert str(excinfo.value) == "Cannot compose empty Mixed"


def test_text_alt_html_and_attachment() -> None:
    t = TextBody("This is the text of an e-mail.")
    h = HTMLBody("<p>This is the <i>text</i> of an <b>e</b>-mail.<p>")
    a = TextAttachment(
        "this,text,attachment", "attachment.csv", content_type="text/csv"
    )
    msg = ((t | h) & a).compose(
        from_="me@here.com",
        to=["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
        subject="Some electronic mail",
    )
    assert email2dict(msg) == {
        "unixfrom": None,
        "headers": {
            "subject": "Some electronic mail",
            "from": [
                {
                    "display_name": "",
                    "address": "me@here.com",
                }
            ],
            "to": [
                {
                    "display_name": "",
                    "address": "you@there.net",
                },
                {
                    "display_name": "Thaddeus Hem",
                    "address": "them@hither.yon",
                },
            ],
            "content-type": {
                "content_type": "multipart/mixed",
                "params": {},
            },
        },
        "preamble": None,
        "content": [
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "multipart/alternative",
                        "params": {},
                    },
                },
                "preamble": None,
                "content": [
                    {
                        "unixfrom": None,
                        "headers": {
                            "content-type": {
                                "content_type": "text/plain",
                                "params": {},
                            },
                        },
                        "preamble": None,
                        "content": "This is the text of an e-mail.\n",
                        "epilogue": None,
                    },
                    {
                        "unixfrom": None,
                        "headers": {
                            "content-type": {
                                "content_type": "text/html",
                                "params": {},
                            },
                        },
                        "preamble": None,
                        "content": (
                            "<p>This is the <i>text</i> of an <b>e</b>-mail.<p>\n"
                        ),
                        "epilogue": None,
                    },
                ],
                "epilogue": None,
            },
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "text/csv",
                        "params": {},
                    },
                    "content-disposition": {
                        "disposition": "attachment",
                        "params": {"filename": "attachment.csv"},
                    },
                },
                "preamble": None,
                "content": "this,text,attachment\n",
                "epilogue": None,
            },
        ],
        "epilogue": None,
    }


def test_related() -> None:
    t = TextBody("This is the text of an e-mail.")
    h = HTMLBody("<p>This is the <i>text</i> of an <b>e</b>-mail.<p>")
    rel = t ^ h
    assert isinstance(rel, Related)
    assert rel == Related([t, h])


def test_xor_related() -> None:
    t1 = TextBody("Part 1")
    t2 = TextBody("Part 2")
    t3 = TextBody("Part 3")
    t4 = TextBody("Part 4")
    rel1 = t1 ^ t2
    rel2 = t3 ^ t4
    rel = rel1 ^ rel2
    assert rel == Related([t1, t2, t3, t4])


def test_related_xor_eq_body() -> None:
    t1 = TextBody("Part 1")
    t2 = TextBody("Part 2")
    t3 = TextBody("Part 3")
    rel = t1 ^ t2
    x = rel
    x ^= t3
    assert x is rel
    assert x == Related([t1, t2, t3])
    assert t3 == TextBody("Part 3")


def test_related_xor_eq_related() -> None:
    t1 = TextBody("Part 1")
    t2 = TextBody("Part 2")
    t3 = TextBody("Part 3")
    t4 = TextBody("Part 4")
    rel1 = t1 ^ t2
    rel2 = t3 ^ t4
    x = rel1
    x ^= rel2
    assert x is rel1
    assert x == Related([t1, t2, t3, t4])
    assert rel2 == Related([t3, t4])


def test_body_xor_eq() -> None:
    t = TextBody("This is the text of an e-mail.")
    h = HTMLBody("<p>This is the <i>text</i> of an <b>e</b>-mail.<p>")
    x: Any = t
    x ^= h
    assert x is not t
    assert isinstance(x, Related)
    assert x == Related([t, h])
    assert t == TextBody("This is the text of an e-mail.")


def test_compose_empty_related() -> None:
    with pytest.raises(ValueError) as excinfo:
        Related([]).compose(
            from_="me@here.com",
            to=["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
            subject="Some electronic mail",
        )
    assert str(excinfo.value) == "Cannot compose empty Related"


def test_text_alt_html_related() -> None:
    # Converted from <https://docs.python.org/3/library/email.examples.html>
    TEXT = (
        "Salut!\n"
        "\n"
        "Cela ressemble à un excellent recipie[1] déjeuner.\n"
        "\n"
        "[1] http://www.yummly.com/recipe/Roasted-Asparagus-Epicurious-203718\n"
        "\n"
        "--Pepé\n"
    )
    t = TextBody(TEXT)
    asparagus_cid = make_msgid()
    HTML = (
        "<html>\n"
        "  <head></head>\n"
        "  <body>\n"
        "    <p>Salut!</p>\n"
        "    <p>Cela ressemble à un excellent\n"
        '        <a href="http://www.yummly.com/recipe/Roasted-Asparagus-'
        'Epicurious-203718">\n'
        "            recipie\n"
        "        </a> déjeuner.\n"
        "    </p>\n"
        f'    <img src="cid:{asparagus_cid[1:-1]}" />\n'
        "  </body>\n"
        "</html>\n"
    )
    h = HTMLBody(HTML)
    IMG = b"\1\2\3\4\5"
    img = BytesAttachment(
        IMG,
        "asparagus.png",
        content_type="image/png",
        inline=True,
        content_id=asparagus_cid,
    )
    msg = (t | (h ^ img)).compose(
        subject="Ayons asperges pour le déjeuner",
        from_=Address("Pepé Le Pew", "pepe@example.com"),
        to=(
            Address("Penelope Pussycat", "penelope@example.com"),
            Address("Fabrette Pussycat", "fabrette@example.com"),
        ),
    )
    assert email2dict(msg) == {
        "unixfrom": None,
        "headers": {
            "subject": "Ayons asperges pour le déjeuner",
            "from": [
                {
                    "display_name": "Pepé Le Pew",
                    "address": "pepe@example.com",
                },
            ],
            "to": [
                {
                    "display_name": "Penelope Pussycat",
                    "address": "penelope@example.com",
                },
                {
                    "display_name": "Fabrette Pussycat",
                    "address": "fabrette@example.com",
                },
            ],
            "content-type": {
                "content_type": "multipart/alternative",
                "params": {},
            },
        },
        "preamble": None,
        "content": [
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "text/plain",
                        "params": {},
                    },
                },
                "preamble": None,
                "content": TEXT,
                "epilogue": None,
            },
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "multipart/related",
                        "params": {"type": "text/html"},
                    },
                },
                "preamble": None,
                "content": [
                    {
                        "unixfrom": None,
                        "headers": {
                            "content-type": {
                                "content_type": "text/html",
                                "params": {},
                            },
                        },
                        "preamble": None,
                        "content": HTML,
                        "epilogue": None,
                    },
                    {
                        "unixfrom": None,
                        "headers": {
                            "content-type": {
                                "content_type": "image/png",
                                "params": {},
                            },
                            "content-disposition": {
                                "disposition": "inline",
                                "params": {"filename": "asparagus.png"},
                            },
                            "content-id": [asparagus_cid],
                        },
                        "preamble": None,
                        "content": IMG,
                        "epilogue": None,
                    },
                ],
                "epilogue": None,
            },
        ],
        "epilogue": None,
    }


def test_img_html_related_start() -> None:
    asparagus_cid = make_msgid()
    html_cid = make_msgid()
    HTML = (
        "<html>\n"
        "  <head></head>\n"
        "  <body>\n"
        "    <p>Salut!</p>\n"
        "    <p>Cela ressemble à un excellent\n"
        '        <a href="http://www.yummly.com/recipe/Roasted-Asparagus-'
        'Epicurious-203718">\n'
        "            recipie\n"
        "        </a> déjeuner.\n"
        "    </p>\n"
        f'    <img src="cid:{asparagus_cid[1:-1]}" />\n'
        "  </body>\n"
        "</html>\n"
    )
    h = HTMLBody(HTML, content_id=html_cid)
    IMG = b"\1\2\3\4\5"
    img = BytesAttachment(
        IMG,
        "asparagus.png",
        content_type="image/png",
        inline=True,
        content_id=asparagus_cid,
    )
    rel = img ^ h
    rel.start = html_cid
    msg = rel.compose(
        subject="Ayons asperges pour le déjeuner",
        from_=Address("Pepé Le Pew", "pepe@example.com"),
        to=(
            Address("Penelope Pussycat", "penelope@example.com"),
            Address("Fabrette Pussycat", "fabrette@example.com"),
        ),
    )
    assert email2dict(msg) == {
        "unixfrom": None,
        "headers": {
            "subject": "Ayons asperges pour le déjeuner",
            "from": [
                {
                    "display_name": "Pepé Le Pew",
                    "address": "pepe@example.com",
                },
            ],
            "to": [
                {
                    "display_name": "Penelope Pussycat",
                    "address": "penelope@example.com",
                },
                {
                    "display_name": "Fabrette Pussycat",
                    "address": "fabrette@example.com",
                },
            ],
            "content-type": {
                "content_type": "multipart/related",
                "params": {"type": "text/html", "start": html_cid},
            },
        },
        "preamble": None,
        "content": [
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "image/png",
                        "params": {},
                    },
                    "content-disposition": {
                        "disposition": "inline",
                        "params": {"filename": "asparagus.png"},
                    },
                    "content-id": [asparagus_cid],
                },
                "preamble": None,
                "content": IMG,
                "epilogue": None,
            },
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "text/html",
                        "params": {},
                    },
                    "content-id": [html_cid],
                },
                "preamble": None,
                "content": HTML,
                "epilogue": None,
            },
        ],
        "epilogue": None,
    }


def test_multipart_content_ids() -> None:
    img_cid = make_msgid()
    img = BytesAttachment(
        b"\1\2\3\4\5", filename="blob.png", content_type="image/png", content_id=img_cid
    )
    text = TextBody("Here is an image:\n")
    HTML = f'<p>Here is an image:</p><img src="cid:{img_cid[1:-1]}"/>\n'
    html = HTMLBody(HTML)
    mixed = text & img
    mixed_cid = make_msgid()
    mixed.content_id = mixed_cid
    related = html ^ img
    related_cid = make_msgid()
    related.content_id = related_cid
    alt = mixed | related
    alt_cid = make_msgid()
    alt.content_id = alt_cid
    msg = alt.compose(
        from_="me@here.com",
        to=["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
        subject="Some electronic mail",
    )
    assert email2dict(msg) == {
        "unixfrom": None,
        "headers": {
            "subject": "Some electronic mail",
            "from": [
                {
                    "display_name": "",
                    "address": "me@here.com",
                }
            ],
            "to": [
                {
                    "display_name": "",
                    "address": "you@there.net",
                },
                {
                    "display_name": "Thaddeus Hem",
                    "address": "them@hither.yon",
                },
            ],
            "content-type": {
                "content_type": "multipart/alternative",
                "params": {},
            },
            "content-id": [alt_cid],
        },
        "preamble": None,
        "content": [
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "multipart/mixed",
                        "params": {},
                    },
                    "content-id": [mixed_cid],
                },
                "preamble": None,
                "content": [
                    {
                        "unixfrom": None,
                        "headers": {
                            "content-type": {
                                "content_type": "text/plain",
                                "params": {},
                            },
                        },
                        "preamble": None,
                        "content": "Here is an image:\n",
                        "epilogue": None,
                    },
                    {
                        "unixfrom": None,
                        "headers": {
                            "content-type": {
                                "content_type": "image/png",
                                "params": {},
                            },
                            "content-disposition": {
                                "disposition": "attachment",
                                "params": {"filename": "blob.png"},
                            },
                            "content-id": [img_cid],
                        },
                        "preamble": None,
                        "content": b"\1\2\3\4\5",
                        "epilogue": None,
                    },
                ],
                "epilogue": None,
            },
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "multipart/related",
                        "params": {"type": "text/html"},
                    },
                    "content-id": [related_cid],
                },
                "preamble": None,
                "content": [
                    {
                        "unixfrom": None,
                        "headers": {
                            "content-type": {
                                "content_type": "text/html",
                                "params": {},
                            },
                        },
                        "preamble": None,
                        "content": HTML,
                        "epilogue": None,
                    },
                    {
                        "unixfrom": None,
                        "headers": {
                            "content-type": {
                                "content_type": "image/png",
                                "params": {},
                            },
                            "content-disposition": {
                                "disposition": "attachment",
                                "params": {"filename": "blob.png"},
                            },
                            "content-id": [img_cid],
                        },
                        "preamble": None,
                        "content": b"\1\2\3\4\5",
                        "epilogue": None,
                    },
                ],
                "epilogue": None,
            },
        ],
        "epilogue": None,
    }


@pytest.mark.parametrize("cls", [Alternative, Mixed, Related])
def test_multipart_mutable_sequences(cls: Type[Multipart]) -> None:
    t1 = TextBody("Part 1")
    t2 = TextBody("Part 2")
    t3 = TextBody("Part 3")
    t4 = TextBody("Part 4")
    seq = cls()
    assert len(seq) == 0
    assert list(seq) == []
    seq.append(t1)
    assert len(seq) == 1
    assert list(seq) == [t1]
    assert seq[0] == t1
    seq.insert(0, t2)
    assert len(seq) == 2
    assert list(seq) == [t2, t1]
    assert seq[0] == t2
    assert seq[1] == t1
    seq[1] = t3
    assert len(seq) == 2
    assert list(seq) == [t2, t3]
    assert seq[0] == t2
    assert seq[1] == t3
    seq.pop(0)
    assert len(seq) == 1
    assert list(seq) == [t3]
    assert seq[0] == t3
    seq.extend([t1, t4])
    assert len(seq) == 3
    assert list(seq) == [t3, t1, t4]
    seq.reverse()
    assert len(seq) == 3
    assert list(seq) == [t4, t1, t3]
    seq.reverse()
    assert len(seq) == 3
    assert list(seq) == [t3, t1, t4]
    assert seq[1:] == cls([t1, t4])
    seq.pop()
    assert len(seq) == 2
    assert list(seq) == [t3, t1]
    seq.remove(t1)
    assert len(seq) == 1
    assert list(seq) == [t3]
    seq += [t1, t2]
    assert len(seq) == 3
    assert list(seq) == [t3, t1, t2]
    del seq[0]
    assert len(seq) == 2
    assert list(seq) == [t1, t2]
    seq[1:] = [t3, t4]
    assert len(seq) == 3
    assert list(seq) == [t1, t3, t4]
    del seq[:2]
    assert len(seq) == 1
    assert list(seq) == [t4]


def test_str_or_html() -> None:
    t = "This is the text of an e-mail."
    h = HTMLBody("<p>This is the <i>text</i> of an <b>e</b>-mail.<p>")
    alt = t | h
    assert isinstance(alt, Alternative)
    assert alt == Alternative([TextBody(t), h])


def test_html_or_str() -> None:
    h = HTMLBody("<p>This is the <i>text</i> of an <b>e</b>-mail.<p>")
    t = "This is the text of an e-mail."
    alt = h | t
    assert isinstance(alt, Alternative)
    assert alt == Alternative([h, TextBody(t)])


def test_alt_or_eq_str() -> None:
    t1 = TextBody("Part 1")
    t2 = TextBody("Part 2")
    t3 = "Part 3"
    alt = t1 | t2
    x = alt
    x |= t3
    assert x is alt
    assert x == Alternative([t1, t2, TextBody(t3)])
    assert t3 == "Part 3"


def test_str_or_eq_body() -> None:
    t = "This is the text of an e-mail."
    h = HTMLBody("<p>This is the <i>text</i> of an <b>e</b>-mail.<p>")
    x: Any = t
    x |= h
    assert x is not t
    assert isinstance(x, Alternative)
    assert x == Alternative([TextBody(t), h])
    assert t == "This is the text of an e-mail."


def test_body_or_eq_str() -> None:
    h = HTMLBody("<p>This is the <i>text</i> of an <b>e</b>-mail.<p>")
    t = "This is the text of an e-mail."
    x: Any = h
    x |= t
    assert x is not h
    assert isinstance(x, Alternative)
    assert x == Alternative([h, TextBody(t)])
    assert h == HTMLBody("<p>This is the <i>text</i> of an <b>e</b>-mail.<p>")
    assert t == "This is the text of an e-mail."


def test_str_and_attachment() -> None:
    t = "This is the text of an e-mail."
    a = TextAttachment(
        "this,text,attachment", "attachment.csv", content_type="text/csv"
    )
    mixed = t & a
    assert isinstance(mixed, Mixed)
    assert mixed == Mixed([TextBody(t), a])


def test_attachment_and_str() -> None:
    a = TextAttachment(
        "this,text,attachment", "attachment.csv", content_type="text/csv"
    )
    t = "This is the text of an e-mail."
    mixed = a & t
    assert isinstance(mixed, Mixed)
    assert mixed == Mixed([a, TextBody(t)])


def test_mixed_and_eq_str() -> None:
    t1 = TextBody("Part 1")
    t2 = TextBody("Part 2")
    t3 = "Part 3"
    mixed = t1 & t2
    x = mixed
    x &= t3
    assert x is mixed
    assert x == Mixed([t1, t2, TextBody(t3)])
    assert t3 == "Part 3"


def test_str_and_eq_body() -> None:
    t = "This is the text of an e-mail."
    h = HTMLBody("<p>This is the <i>text</i> of an <b>e</b>-mail.<p>")
    x: Any = t
    x &= h
    assert x is not t
    assert isinstance(x, Mixed)
    assert x == Mixed([TextBody(t), h])
    assert t == "This is the text of an e-mail."


def test_body_and_eq_str() -> None:
    h = HTMLBody("<p>This is the <i>text</i> of an <b>e</b>-mail.<p>")
    t = "This is the text of an e-mail."
    x: Any = h
    x &= t
    assert x is not h
    assert isinstance(x, Mixed)
    assert x == Mixed([h, TextBody(t)])
    assert h == HTMLBody("<p>This is the <i>text</i> of an <b>e</b>-mail.<p>")
    assert t == "This is the text of an e-mail."


def test_str_xor_body() -> None:
    t = "This is the text of an e-mail."
    h = HTMLBody("<p>This is the <i>text</i> of an <b>e</b>-mail.<p>")
    rel = t ^ h
    assert isinstance(rel, Related)
    assert rel == Related([TextBody(t), h])


def test_body_xor_str() -> None:
    h = HTMLBody("<p>This is the <i>text</i> of an <b>e</b>-mail.<p>")
    t = "This is the text of an e-mail."
    rel = h ^ t
    assert isinstance(rel, Related)
    assert rel == Related([h, TextBody(t)])


def test_related_xor_eq_str() -> None:
    t1 = TextBody("Part 1")
    t2 = TextBody("Part 2")
    t3 = "Part 3"
    rel = t1 ^ t2
    x = rel
    x ^= t3
    assert x is rel
    assert x == Related([t1, t2, TextBody(t3)])
    assert t3 == "Part 3"


def test_str_xor_eq_body() -> None:
    t = "This is the text of an e-mail."
    h = HTMLBody("<p>This is the <i>text</i> of an <b>e</b>-mail.<p>")
    x: Any = t
    x ^= h
    assert x is not t
    assert isinstance(x, Related)
    assert x == Related([TextBody(t), h])
    assert t == "This is the text of an e-mail."


def test_body_xor_eq_str() -> None:
    h = HTMLBody("<p>This is the <i>text</i> of an <b>e</b>-mail.<p>")
    t = "This is the text of an e-mail."
    x: Any = h
    x ^= t
    assert x is not h
    assert isinstance(x, Related)
    assert x == Related([h, TextBody(t)])
    assert h == HTMLBody("<p>This is the <i>text</i> of an <b>e</b>-mail.<p>")
    assert t == "This is the text of an e-mail."



from collections.abc import Iterable as IterableABC
from email import headerregistry as hr
from mimetypes import guess_type
import os
from typing import Iterable, List, Union

AnyPath = Union[bytes, str, "os.PathLike[bytes]", "os.PathLike[str]"]

SingleAddress = Union[str, hr.Address]

AddressOrGroup = Union[str, hr.Address, hr.Group]


def compile_address(addr: SingleAddress) -> hr.Address:
    if isinstance(addr, str):
        return hr.Address(addr_spec=addr)
    else:
        return addr


def compile_addresses(
    addrs: Union[AddressOrGroup, Iterable[AddressOrGroup]]
) -> List[Union[hr.Address, hr.Group]]:
    if isinstance(addrs, str):
        return [hr.Address(addr_spec=addrs)]
    elif not isinstance(addrs, IterableABC):
        return [addrs]
    else:
        return [hr.Address(addr_spec=a) if isinstance(a, str) else a for a in addrs]


def get_mime_type(filename: str, strict: bool = False) -> str:
    """
    Like `mimetypes.guess_type()`, except that if the file is compressed, the
    MIME type for the compression is returned.  Also, the default return value
    is now ``'application/octet-stream'`` instead of `None`.
    """
    mtype, encoding = guess_type(filename, strict)
    if encoding is None:
        return mtype or "application/octet-stream"
    elif encoding == "gzip":
        # application/gzip is defined by RFC 6713
        return "application/gzip"
        # There is also a "+gzip" MIME structured syntax suffix defined by RFC
        # 8460; exactly when can that be used?
        # return mtype + '+gzip'
    else:
        return "application/x-" + encoding



from typing import Dict
import pytest
from eletter import assemble_content_type, reply_quote
from eletter.util import get_mime_type


@pytest.mark.parametrize(
    "maintype,subtype,params,ct",
    [
        ("text", "plain", {}, "text/plain"),
        ("TEXT", "PLAIN", {}, "TEXT/PLAIN"),
        ("text", "plain", {"charset": "utf-8"}, 'text/plain; charset="utf-8"'),
        ("text", "plain", {"name": "résumé.txt"}, 'text/plain; name="résumé.txt"'),
        ("text", "plain", {"name": 'foo"bar'}, 'text/plain; name="foo\\"bar"'),
        (
            "text",
            "markdown",
            {"charset": "utf-8", "variant": "GFM"},
            'text/markdown; charset="utf-8"; variant="GFM"',
        ),
    ],
)
def test_assemble_content_type(
    maintype: str, subtype: str, params: Dict[str, str], ct: str
) -> None:
    assert assemble_content_type(maintype, subtype, **params) == ct


@pytest.mark.parametrize(
    "maintype,subtype",
    [
        ("text/plain", "plain"),
        ("text", ""),
        ("text/", "plain"),
    ],
)
def test_assemble_content_type_error(maintype: str, subtype: str) -> None:
    with pytest.raises(ValueError) as excinfo:
        assemble_content_type(maintype, subtype)
    assert str(excinfo.value) == f"{maintype}/{subtype}"


@pytest.mark.parametrize(
    "filename,mtype",
    [
        ("foo.txt", "text/plain"),
        ("foo", "application/octet-stream"),
        ("foo.gz", "application/gzip"),
        ("foo.tar.gz", "application/gzip"),
        ("foo.tgz", "application/gzip"),
        ("foo.taz", "application/gzip"),
        ("foo.svg.gz", "application/gzip"),
        ("foo.svgz", "application/gzip"),
        ("foo.Z", "application/x-compress"),
        ("foo.tar.Z", "application/x-compress"),
        ("foo.bz2", "application/x-bzip2"),
        ("foo.tar.bz2", "application/x-bzip2"),
        ("foo.tbz2", "application/x-bzip2"),
        ("foo.xz", "application/x-xz"),
        ("foo.tar.xz", "application/x-xz"),
        ("foo.txz", "application/x-xz"),
    ],
)
def test_get_mime_type(filename: str, mtype: str) -> None:
    assert get_mime_type(filename) == mtype


@pytest.mark.parametrize(
    "inp,output",
    [
        ("", "> \n"),
        ("\n", "> \n"),
        ("Insert output here.", "> Insert output here.\n"),
        ("Insert output here.\n", "> Insert output here.\n"),
        (
            "Insert output here.\nOutsert input there.",
            "> Insert output here.\n> Outsert input there.\n",
        ),
        (
            "Insert output here.\nOutsert input there.\n",
            "> Insert output here.\n> Outsert input there.\n",
        ),
        (
            "Insert output here.\r\nOutsert input there.\r\n",
            "> Insert output here.\r\n> Outsert input there.\r\n",
        ),
        (
            "Insert output here.\rOutsert input there.\r",
            "> Insert output here.\r> Outsert input there.\r",
        ),
        (
            "Insert output here.\n\nOutsert input there.\n",
            "> Insert output here.\n> \n> Outsert input there.\n",
        ),
        (
            "> Insert output here.\n> \n> Outsert input there.\n",
            ">> Insert output here.\n>> \n>> Outsert input there.\n",
        ),
    ],
)
def test_reply_quote(inp: str, output: str) -> None:
    assert reply_quote(inp) == output


def test_reply_quote_custom_prefix() -> None:
    assert (
        reply_quote("Insert output here.\n\n: Outsert input there.\n", ": ")
        == ": Insert output here.\n: \n:: Outsert input there.\n"
    )



from abc import ABC, abstractmethod
from collections.abc import Iterable as IterableABC
from datetime import datetime
from email import headerregistry as hr
from email import message_from_binary_file
from email import policy
from email.message import EmailMessage
import os
import os.path
from typing import (
    Iterable,
    List,
    Mapping,
    MutableSequence,
    Optional,
    TypeVar,
    Union,
    overload,
)
import attr
from mailbits import ContentType
from .util import (
    AddressOrGroup,
    AnyPath,
    SingleAddress,
    compile_address,
    compile_addresses,
    get_mime_type,
)


class Address(hr.Address):
    """ A combination of a person's name and their e-mail address """

    def __init__(self, display_name: str, address: str) -> None:
        super().__init__(display_name=display_name, addr_spec=address)


class Group(hr.Group):
    """ An e-mail address group """

    def __init__(self, display_name: str, addresses: Iterable[SingleAddress]) -> None:
        super().__init__(
            display_name=display_name, addresses=tuple(map(compile_address, addresses))
        )


def cache_content_type(
    ctd: "ContentTyped",
    _attr: Optional[attr.Attribute],
    value: str,
) -> None:
    ct = ContentType.parse(value)
    if ctd.DEFAULT_CONTENT_TYPE.startswith("text/") and ct.maintype != "text":
        raise ValueError("content_type must be text/*")
    ctd._ct = ct


@attr.s
class ContentTyped:
    DEFAULT_CONTENT_TYPE = "application/octet-stream"

    #: The :mailheader:`Content-Type` of the attachment
    content_type: str = attr.ib(
        kw_only=True,
        default=attr.Factory(lambda self: self.DEFAULT_CONTENT_TYPE, takes_self=True),
        on_setattr=cache_content_type,
    )
    _ct: ContentType = attr.ib(init=False, repr=False, eq=False, order=False)

    def __attrs_post_init__(self) -> None:
        cache_content_type(self, None, self.content_type)


@attr.s
class MailItem(ABC):
    """ Base class for all ``eletter`` message components """

    #: :mailheader:`Content-ID` header value for the item
    content_id: Optional[str] = attr.ib(default=None, kw_only=True)

    @abstractmethod
    def _compile(self) -> EmailMessage:
        ...

    def compose(
        self,
        *,
        subject: str,
        to: Iterable[AddressOrGroup],
        from_: Optional[Union[AddressOrGroup, Iterable[AddressOrGroup]]] = None,
        cc: Optional[Iterable[AddressOrGroup]] = None,
        bcc: Optional[Iterable[AddressOrGroup]] = None,
        reply_to: Optional[Union[AddressOrGroup, Iterable[AddressOrGroup]]] = None,
        sender: Optional[SingleAddress] = None,
        date: Optional[datetime] = None,
        headers: Optional[Mapping[str, Union[str, Iterable[str]]]] = None,
    ) -> EmailMessage:
        """
        Convert the `MailItem` into an `~email.message.EmailMessage` with the
        item's contents as the payload and with the given subject,
        :mailheader:`From` address, :mailheader:`To` addresses, and optional
        other headers.

        All parameters other than ``subject`` and ``to`` are optional.

        :param str subject: The e-mail's :mailheader:`Subject` line
        :param to: The e-mail's :mailheader:`To` line
        :type to: iterable of addresses
        :param from_:
            The e-mail's :mailheader:`From` line.  Note that this argument is
            spelled with an underscore, as "``from``" is a keyword in Python.
        :type from_: address or iterable of addresses
        :param cc: The e-mail's :mailheader:`CC` line
        :type cc: iterable of addresses
        :param bcc: The e-mail's :mailheader:`BCC` line
        :type bcc: iterable of addresses
        :param reply_to: The e-mail's :mailheader:`Reply-To` line
        :type reply_to: address or iterable of addresses
        :param address sender:
            The e-mail's :mailheader:`Sender` line.  The address must be a
            string or `Address`, not a `Group`.
        :param datetime date: The e-mail's :mailheader:`Date` line
        :param mapping headers:
            A collection of additional headers to add to the e-mail.  A header
            value may be either a single string or an iterable of strings to add
            multiple headers with the same name.  If you wish to set an
            otherwise-unsupported address header like :mailheader:`Resent-From`
            to a list of addresses, use the `format_addresses()` function to
            first convert the addresses to a string.
        :rtype: email.message.EmailMessage
        """
        msg = self._compile()
        msg["Subject"] = subject
        msg["To"] = compile_addresses(to)
        if from_ is not None:
            msg["From"] = compile_addresses(from_)
        if cc is not None:
            msg["CC"] = compile_addresses(cc)
        if bcc is not None:
            msg["BCC"] = compile_addresses(bcc)
        if reply_to is not None:
            msg["Reply-To"] = compile_addresses(reply_to)
        if sender is not None:
            msg["Sender"] = compile_address(sender)
        if date is not None:
            msg["Date"] = date
        if headers is not None:
            for k, v in headers.items():
                values: List[str]
                if isinstance(v, str):
                    values = [v]
                else:
                    values = list(v)
                for v2 in values:
                    msg[k] = v2
        return msg

    def __or__(self, other: Union["MailItem", str]) -> "Alternative":
        parts: List[MailItem] = []
        for mi in [self, other]:
            if isinstance(mi, Alternative):
                parts.extend(mi.content)
            elif isinstance(mi, str):
                parts.append(TextBody(mi))
            else:
                parts.append(mi)
        return Alternative(parts)

    def __ror__(self, other: Union["MailItem", str]) -> "Alternative":
        parts: List[MailItem] = []
        for mi in [other, self]:
            if isinstance(mi, Alternative):
                parts.extend(mi.content)  # pragma: no cover
            elif isinstance(mi, str):
                parts.append(TextBody(mi))
            else:
                parts.append(mi)  # pragma: no cover
        return Alternative(parts)

    def __and__(self, other: Union["MailItem", str]) -> "Mixed":
        parts: List[MailItem] = []
        for mi in [self, other]:
            if isinstance(mi, Mixed):
                parts.extend(mi.content)
            elif isinstance(mi, str):
                parts.append(TextBody(mi))
            else:
                parts.append(mi)
        return Mixed(parts)

    def __rand__(self, other: Union["MailItem", str]) -> "Mixed":
        parts: List[MailItem] = []
        for mi in [other, self]:
            if isinstance(mi, Mixed):
                parts.extend(mi.content)  # pragma: no cover
            elif isinstance(mi, str):
                parts.append(TextBody(mi))
            else:
                parts.append(mi)  # pragma: no cover
        return Mixed(parts)

    def __xor__(self, other: Union["MailItem", str]) -> "Related":
        parts: List[MailItem] = []
        for mi in [self, other]:
            if isinstance(mi, Related):
                parts.extend(mi.content)
            elif isinstance(mi, str):
                parts.append(TextBody(mi))
            else:
                parts.append(mi)
        return Related(parts)

    def __rxor__(self, other: Union["MailItem", str]) -> "Related":
        parts: List[MailItem] = []
        for mi in [other, self]:
            if isinstance(mi, Related):
                parts.extend(mi.content)  # pragma: no cover
            elif isinstance(mi, str):
                parts.append(TextBody(mi))
            else:
                parts.append(mi)  # pragma: no cover
        return Related(parts)


class Attachment(MailItem):
    """ Base class for the attachment classes """

    pass


@attr.s(auto_attribs=True)
class TextAttachment(Attachment, ContentTyped):
    """
    A textual e-mail attachment.  ``content_type`` defaults to ``"text/plain"``
    and must have a maintype of :mimetype:`text`.
    """

    DEFAULT_CONTENT_TYPE = "text/plain"

    #: The body of the attachment
    content: str
    #: The filename of the attachment
    filename: str
    #: Whether the attachment should be displayed inline in clients
    inline: bool = attr.ib(default=False, kw_only=True)

    def _compile(self) -> EmailMessage:
        assert self._ct.maintype == "text", "Content-Type is not text/*"
        params = dict(self._ct.params)
        charset = params.pop("charset", "utf-8")
        msg = EmailMessage()
        msg.set_content(
            self.content,
            subtype=self._ct.subtype,
            disposition="inline" if self.inline else "attachment",
            filename=self.filename,
            params=params,
            charset=charset,
            cid=self.content_id,
        )
        return msg

    @classmethod
    def from_file(
        cls,
        path: AnyPath,
        content_type: Optional[str] = None,
        encoding: Optional[str] = None,
        errors: Optional[str] = None,
        inline: bool = False,
        content_id: Optional[str] = None,
    ) -> "TextAttachment":
        """
        Construct a `TextAttachment` from the contents of the file at ``path``.
        The filename of the attachment will be set to the basename of ``path``.
        If ``content_type`` is `None`, the :mailheader:`Content-Type` is
        guessed based on ``path``'s file extension.  ``encoding`` and
        ``errors`` are used when opening the file and have no relation to the
        :mailheader:`Content-Type`.
        """
        with open(path, "r", encoding=encoding, errors=errors) as fp:
            content = fp.read()
        filename = os.path.basename(os.fsdecode(path))
        if content_type is None:
            content_type = get_mime_type(filename)
        return cls(
            content=content,
            filename=filename,
            content_type=content_type,
            inline=inline,
            content_id=content_id,
        )


@attr.s(auto_attribs=True)
class BytesAttachment(Attachment, ContentTyped):
    """
    A binary e-mail attachment.  `content_type` defaults to
    ``"application/octet-stream"``.
    """

    #: The body of the attachment
    content: bytes
    #: The filename of the attachment
    filename: str
    #: Whether the attachment should be displayed inline in clients
    inline: bool = attr.ib(default=False, kw_only=True)

    def _compile(self) -> EmailMessage:
        msg = EmailMessage()
        msg.set_content(
            self.content,
            self._ct.maintype,
            self._ct.subtype,
            disposition="inline" if self.inline else "attachment",
            filename=self.filename,
            params=self._ct.params,
            cid=self.content_id,
        )
        return msg

    @classmethod
    def from_file(
        cls,
        path: AnyPath,
        content_type: Optional[str] = None,
        inline: bool = False,
        content_id: Optional[str] = None,
    ) -> "BytesAttachment":
        """
        Construct a `BytesAttachment` from the contents of the file at
        ``path``.  The filename of the attachment will be set to the basename
        of ``path``.  If ``content_type`` is `None`, the
        :mailheader:`Content-Type` is guessed based on ``path``'s file
        extension.
        """
        with open(path, "rb") as fp:
            content = fp.read()
        filename = os.path.basename(os.fsdecode(path))
        if content_type is None:
            content_type = get_mime_type(filename)
        return cls(
            content=content,
            filename=filename,
            content_type=content_type,
            inline=inline,
            content_id=content_id,
        )


@attr.s(auto_attribs=True)
class EmailAttachment(Attachment):
    """ A :mimetype:`message/rfc822` e-mail attachment """

    #: The body of the attachment
    content: EmailMessage
    #: The filename of the attachment
    filename: str
    #: Whether the attachment should be displayed inline in clients
    inline: bool = attr.ib(default=False, kw_only=True)

    def _compile(self) -> EmailMessage:
        msg = EmailMessage()
        msg.set_content(
            self.content,
            disposition="inline" if self.inline else "attachment",
            filename=self.filename,
            cid=self.content_id,
        )
        return msg

    @classmethod
    def from_file(
        cls, path: AnyPath, inline: bool = False, content_id: Optional[str] = None
    ) -> "EmailAttachment":
        """
        Construct an `EmailAttachment` from the contents of the file at
        ``path``.  The filename of the attachment will be set to the basename
        of ``path``.
        """
        with open(path, "rb") as fp:
            content = message_from_binary_file(fp, policy=policy.default)
            assert isinstance(content, EmailMessage)
        filename = os.path.basename(os.fsdecode(path))
        return cls(
            content=content, filename=filename, inline=inline, content_id=content_id
        )


def mail_item_list(xs: Iterable[MailItem]) -> List[MailItem]:
    return list(xs)


M = TypeVar("M", bound="Multipart")


@attr.s
class Multipart(MailItem, MutableSequence[MailItem]):
    """
    Base class for all multipart classes.  All such classes are mutable
    sequences of `MailItem`\\s supporting the usual methods (construction from
    an iterable, subscription, ``append()``, ``pop()``, etc.).
    """

    #: The `MailItem`\s contained within the instance
    content: List[MailItem] = attr.ib(factory=list, converter=mail_item_list)

    @overload
    def __getitem__(self, index: int) -> MailItem:
        ...

    @overload
    def __getitem__(self: M, index: slice) -> M:
        ...

    def __getitem__(self: M, index: Union[int, slice]) -> Union[MailItem, M]:
        if isinstance(index, int):
            return self.content[index]
        else:
            return type(self)(self.content[index])

    @overload
    def __setitem__(self, index: int, mi: MailItem) -> None:
        ...

    @overload
    def __setitem__(self, index: slice, mis: Iterable[MailItem]) -> None:
        ...

    def __setitem__(
        self, index: Union[int, slice], mi: Union[MailItem, Iterable[MailItem]]
    ) -> None:
        if isinstance(index, int):
            assert isinstance(mi, MailItem)
            self.content[index] = mi
        else:
            assert isinstance(mi, IterableABC)
            self.content[index] = mi

    def __delitem__(self: M, index: Union[int, slice]) -> None:
        del self.content[index]

    def __len__(self) -> int:
        return len(self.content)

    def insert(self, index: int, value: MailItem) -> None:
        """ :meta private: """
        self.content.insert(index, value)

    def append(self, value: MailItem) -> None:
        """ :meta private: """
        self.content.append(value)

    def reverse(self) -> None:
        """ :meta private: """
        self.content.reverse()

    def extend(self, values: Iterable[MailItem]) -> None:
        """ :meta private: """
        self.content.extend(values)

    def pop(self, index: int = -1) -> MailItem:
        """ :meta private: """
        return self.content.pop(index)

    def remove(self, value: MailItem) -> None:
        """ :meta private: """
        self.content.remove(value)

    def __iadd__(self: M, other: Iterable[MailItem]) -> M:
        self.content.extend(other)
        return self


@attr.s
class Alternative(Multipart):
    """
    A :mimetype:`multipart/alternative` e-mail payload.  E-mails clients will
    display the resulting payload by choosing whichever part they support best.

    An `Alternative` instance can be created by combining two or more
    `MailItem`\\s with the ``|`` operator:

    .. code:: python

        text = TextBody("This is displayed on plain text clients.\\n")
        html = HTMLBody("<p>This is displayed on graphical clients.<p>\\n")

        alternative = text | html

    Likewise, additional `MailItem`\\s can be added to an `Alternative`
    instance with the ``|=`` operator:

    .. code:: python

        # Same as above:
        alternative = Alternative()
        alternative |= TextBody("This is displayed on plain text clients.\\n")
        alternative |= HTMLBody("<p>This is displayed on graphical clients.<p>\\n")

    Using ``|`` to combine a `MailItem` with a `str` automatically converts the
    `str` to a `TextBody`:

    .. code:: python

        # Same as above:

        text = "This is displayed on plain text clients.\\n"
        html = HTMLBody("<p>This is displayed on graphical clients.<p>\\n")

        alternative = text | html

        assert alternative.contents == [
            TextBody("This is displayed on plain text clients.\\n"),
            HTMLBody("<p>This is displayed on graphical clients.<p>\\n"),
        ]

    When combining two `Alternative` instances with ``|`` or ``|=``, the
    contents are "flattened":

    .. code:: python

        # Same as above:
        txtalt = Alternative([
            TextBody("This is displayed on plain text clients.\\n")
        ])
        htmlalt = Alternative([
            HTMLBody("<p>This is displayed on graphical clients.<p>\\n")
        ])
        alternative = txtalt | htmlalt
        assert alternative.contents == [
            TextBody("This is displayed on plain text clients.\\n"),
            HTMLBody("<p>This is displayed on graphical clients.<p>\\n"),
        ]
    """

    def _compile(self) -> EmailMessage:
        if not self.content:
            raise ValueError("Cannot compose empty Alternative")
        msg = EmailMessage()
        msg.make_alternative()
        for mi in self.content:
            msg.attach(mi._compile())
        if self.content_id is not None:
            msg["Content-ID"] = self.content_id
        return msg

    def __ior__(self, other: Union[MailItem, str]) -> "Alternative":
        if isinstance(other, Alternative):
            self.content.extend(other.content)
        elif isinstance(other, str):
            self.content.append(TextBody(other))
        else:
            self.content.append(other)
        return self


@attr.s
class Mixed(Multipart):
    """
    A :mimetype:`multipart/mixed` e-mail payload.  E-mails clients will
    display the resulting payload one part after another, with attachments
    displayed inline if their ``inline`` attribute is set.

    A `Mixed` instance can be created by combining two or more `MailItem`\\s
    with the ``&`` operator:

    .. code:: python

        text = TextBody("Look at the pretty kitty!\\n")
        image = BytesAttachment.from_file("snuffles.jpeg", inline=True)
        sig = TextBody("Sincerely, Me\\n")

        mixed = text & image & sig

    Likewise, additional `MailItem`\\s can be added to a `Mixed` instance with
    the ``&=`` operator:

    .. code:: python

        # Same as above:
        mixed = Mixed()
        mixed &= TextBody("Look at the pretty kitty!\\n")
        mixed &= BytesAttachment.from_file("snuffles.jpeg", inline=True)
        mixed &= TextBody("Sincerely, Me\\n")

    Using ``&`` to combine a `MailItem` with a `str` automatically converts the
    `str` to a `TextBody`:

    .. code:: python

        # Same as above:
        image = BytesAttachment.from_file("snuffles.jpeg", inline=True)

        mixed = "Look at the pretty kitty!\\n" & image & "Sincerely, Me\\n"

        assert mixed.contents == [
            TextBody("Look at the pretty kitty!\\n"),
            BytesAttachment.from_file("snuffles.jpeg", inline=True),
            TextBody("Sincerely, Me\\n"),
        ]

    When combining two `Mixed` instances with ``&`` or ``&=``, the contents are
    "flattened":

    .. code:: python

        part1 = Mixed()
        part1 &= TextBody("Look at the pretty kitty!\\n")
        part1 &= BytesAttachment.from_file("snuffles.jpeg", inline=True)

        part2 = Mixed()
        part2 &= TextBody("Now look at this dog.\\n")
        part2 &= BytesAttachment.from_file("rags.jpeg", inline=True)
        part2 &= TextBody("Which one is cuter?\\n")

        mixed = part1 & part2

        assert mixed.contents == [
            TextBody("Look at the pretty kitty!\\n"),
            BytesAttachment.from_file("snuffles.jpeg", inline=True),
            TextBody("Now look at this dog.\\n"),
            BytesAttachment.from_file("rags.jpeg", inline=True),
            TextBody("Which one is cuter?\\n"),
        ]
    """

    def _compile(self) -> EmailMessage:
        if not self.content:
            raise ValueError("Cannot compose empty Mixed")
        msg = EmailMessage()
        msg.make_mixed()
        for mi in self.content:
            msg.attach(mi._compile())
        if self.content_id is not None:
            msg["Content-ID"] = self.content_id
        return msg

    def __iand__(self, other: Union[MailItem, str]) -> "Mixed":
        if isinstance(other, Mixed):
            self.content.extend(other.content)
        elif isinstance(other, str):
            self.content.append(TextBody(other))
        else:
            self.content.append(other)
        return self


@attr.s(auto_attribs=True)
class Related(Multipart):
    """
    A :mimetype:`multipart/related` e-mail payload.  E-mail clients will
    display the part indicated by the `start` parameter, or the first part if
    `start` is not set.  This part may refer to other parts (e.g., images or
    CSS stylesheets) by their :mailheader:`Content-ID` headers, which can be
    generated using `email.utils.make_msgid()`.

    .. note::

        :mailheader:`Content-ID` headers begin & end with angle brackets
        (``<...>``), which need to be stripped off before including the ID in
        the starting part.

    A `Related` instance can be created by combining two or more `MailItem`\\s
    with the ``^`` operator:

    .. code:: python

        from email.utils import make_msgid

        img_cid = make_msgid()

        html = HTMLBody(
            "<p>Look at the pretty kitty!</p>"
            f'<img src="cid:{img_cid[1:-1]}"/>"
            "<p>Isn't he <em>darling</em>?</p>"
        )

        image = BytesAttachment.from_file("snuffles.jpeg", content_id=img_cid)

        related = html ^ image

    Likewise, additional `MailItem`\\s can be added to a `Related` instance
    with the ``^=`` operator:

    .. code:: python

        # Same as above:

        img_cid = make_msgid()

        related = Related()

        related ^= HTMLBody(
            "<p>Look at the pretty kitty!</p>"
            f'<img src="cid:{img_cid[1:-1]}"/>"
            "<p>Isn't he <em>darling</em>?</p>"
        )

        related ^= BytesAttachment.from_file("snuffles.jpeg", content_id=img_cid)

    Using ``^`` to combine a `MailItem` with a `str` automatically converts the
    `str` to a `TextBody`, though this is generally not all that useful, as
    you'll usually want to create `Related` instances from `HTMLBody`\\s
    instead.

    When combining two `Related` instances with ``^`` or ``^=``, the contents
    are "flattened":

    .. code:: python

        # Same as above:

        img_cid = make_msgid()

        htmlrel = Related([
            HTMLBody(
                "<p>Look at the pretty kitty!</p>"
                f'<img src="cid:{img_cid[1:-1]}"/>"
                "<p>Isn't he <em>darling</em>?</p>"
            )
        ])

        imgrel = Related([
            BytesAttachment.from_file("snuffles.jpeg", content_id=img_cid)
        ])

        related = htmlrel ^ imgrel

        assert related.contents == [
            HTMLBody(
                "<p>Look at the pretty kitty!</p>"
                f'<img src="cid:{img_cid[1:-1]}"/>"
                "<p>Isn't he <em>darling</em>?</p>"
            ),
            BytesAttachment.from_file("snuffles.jpeg", content_id=img_cid),
        ]
    """

    #: The :mailheader:`Content-ID` of the part to display (defaults to the
    #: first part)
    start: Optional[str] = None

    def _compile(self) -> EmailMessage:
        if not self.content:
            raise ValueError("Cannot compose empty Related")
        msg = EmailMessage()
        msg.make_related()
        ctype: Optional[str] = None
        for mi in self.content:
            obj = mi._compile()
            if self.start is not None and mi.content_id == self.start:
                ctype = obj.get_content_type()
            msg.attach(obj)
        if self.content_id is not None:
            msg["Content-ID"] = self.content_id
        if ctype is None:
            ctype = msg.get_payload()[0].get_content_type()
        msg.set_param("type", ctype)
        if self.start is not None:
            msg.set_param("start", self.start)
        return msg

    def __ixor__(self, other: Union[MailItem, str]) -> "Related":
        if isinstance(other, Related):
            self.content.extend(other.content)
        elif isinstance(other, str):
            self.content.append(TextBody(other))
        else:
            self.content.append(other)
        return self


@attr.s(auto_attribs=True)
class TextBody(MailItem):
    """ A :mimetype:`text/plain` e-mail body """

    #: The plain text body
    content: str

    def _compile(self) -> EmailMessage:
        msg = EmailMessage()
        msg.set_content(self.content, cid=self.content_id)
        return msg


@attr.s(auto_attribs=True)
class HTMLBody(MailItem):
    """ A :mimetype:`text/html` e-mail body """

    #: The HTML source of the body
    content: str

    def _compile(self) -> EmailMessage:
        msg = EmailMessage()
        msg.set_content(self.content, subtype="html", cid=self.content_id)
        return msg



from pathlib import Path
from typing import Optional
import attr
from mailbits import ContentType, email2dict
import pytest
from eletter import BytesAttachment, EmailAttachment, TextAttachment

DATA_DIR = Path(__file__).with_name("data")

PNG = bytes.fromhex(
    "89 50 4e 47 0d 0a 1a 0a  00 00 00 0d 49 48 44 52"
    "00 00 00 10 00 00 00 10  08 06 00 00 00 1f f3 ff"
    "61 00 00 00 06 62 4b 47  44 00 ff 00 ff 00 ff a0"
    "bd a7 93 00 00 00 09 70  48 59 73 00 00 00 48 00"
    "00 00 48 00 46 c9 6b 3e  00 00 00 09 76 70 41 67"
    "00 00 00 10 00 00 00 10  00 5c c6 ad c3 00 00 00"
    "5b 49 44 41 54 38 cb c5  92 51 0a c0 30 08 43 7d"
    "b2 fb 5f 39 fb 12 da 61  a9 c3 8e f9 a7 98 98 48"
    "90 64 9d f2 16 da cc ae  b1 01 26 39 92 d8 11 10"
    "16 9e e0 8c 64 dc 89 b9  67 80 ca e5 f3 3f a8 5c"
    "cd 76 52 05 e1 b5 42 ea  1d f0 91 1f b4 09 78 13"
    "e5 52 0e 00 ad 42 f5 bf  85 4f 14 dc 46 b3 32 11"
    "6c b1 43 99 00 00 00 00  49 45 4e 44 ae 42 60 82"
)

PY = """\
def fibonacci(n):
    (a, b) = (0, 1)
    for _ in range(n):
        (a, b) = (b, a + b)
    return a
"""


def test_text_attachment_non_text_content_type() -> None:
    with pytest.raises(ValueError) as excinfo:
        TextAttachment("[]", filename="foo.json", content_type="application/json")
    assert str(excinfo.value) == "content_type must be text/*"


def test_text_attachment_set_non_text_content_type() -> None:
    a = TextAttachment("[]", filename="foo.json")
    with pytest.raises(ValueError) as excinfo:
        a.content_type = "application/json"
    assert str(excinfo.value) == "content_type must be text/*"
    assert a.content_type == "text/plain"
    assert a._ct == ContentType("text", "plain", {})


def test_text_attachment_bad_content_type() -> None:
    with pytest.raises(ValueError) as excinfo:
        TextAttachment("[]", filename="foo.json", content_type="application-json")
    assert str(excinfo.value) == "application-json"


def test_text_attachment_set_bad_content_type() -> None:
    a = TextAttachment("[]", filename="foo.json")
    with pytest.raises(ValueError) as excinfo:
        a.content_type = "application-json"
    assert str(excinfo.value) == "application-json"
    assert a.content_type == "text/plain"
    assert a._ct == ContentType("text", "plain", {})


def test_bytes_attachment_bad_content_type() -> None:
    with pytest.raises(ValueError) as excinfo:
        BytesAttachment(b"[]", filename="foo.json", content_type="application-json")
    assert str(excinfo.value) == "application-json"


def test_bytes_attachment_set_bad_content_type() -> None:
    a = BytesAttachment(b"[]", filename="foo.json")
    with pytest.raises(ValueError) as excinfo:
        a.content_type = "application-json"
    assert str(excinfo.value) == "application-json"
    assert a.content_type == "application/octet-stream"
    assert a._ct == ContentType("application", "octet-stream", {})


def test_bytes_attachment_from_file() -> None:
    ba = BytesAttachment.from_file(DATA_DIR / "ternary.png")
    assert ba == BytesAttachment(
        PNG, filename="ternary.png", content_type="image/png", inline=False
    )


def test_bytes_attachment_from_file_content_type() -> None:
    ba = BytesAttachment.from_file(
        DATA_DIR / "ternary.png", content_type="application/x-ternary"
    )
    assert ba == BytesAttachment(
        PNG, filename="ternary.png", content_type="application/x-ternary", inline=False
    )


def test_bytes_attachment_from_file_inline() -> None:
    ba = BytesAttachment.from_file(DATA_DIR / "ternary.png", inline=True)
    assert ba == BytesAttachment(
        PNG, filename="ternary.png", content_type="image/png", inline=True
    )


def test_bytes_attachment_from_file_content_id() -> None:
    ba = BytesAttachment.from_file(DATA_DIR / "ternary.png", content_id="cid")
    assert ba == BytesAttachment(
        PNG,
        filename="ternary.png",
        content_type="image/png",
        inline=False,
        content_id="cid",
    )


def test_text_attachment_from_file() -> None:
    ta = TextAttachment.from_file(DATA_DIR / "fibonacci.py")
    assert ta == TextAttachment(
        PY, filename="fibonacci.py", content_type="text/x-python", inline=False
    )


def test_text_attachment_from_file_content_type() -> None:
    ta = TextAttachment.from_file(
        DATA_DIR / "fibonacci.py", content_type="text/x-fibonacci; lang=python"
    )
    assert ta == TextAttachment(
        PY,
        filename="fibonacci.py",
        content_type="text/x-fibonacci; lang=python",
        inline=False,
    )


def test_text_attachment_from_file_inline() -> None:
    ta = TextAttachment.from_file(DATA_DIR / "fibonacci.py", inline=True)
    assert ta == TextAttachment(
        PY, filename="fibonacci.py", content_type="text/x-python", inline=True
    )


def test_text_attachment_from_file_content_id() -> None:
    ta = TextAttachment.from_file(DATA_DIR / "fibonacci.py", content_id="cid")
    assert ta == TextAttachment(
        PY,
        filename="fibonacci.py",
        content_type="text/x-python",
        inline=False,
        content_id="cid",
    )


@pytest.mark.parametrize("inline", [False, True])
@pytest.mark.parametrize("cid", [None, "cid"])
def test_email_attachment_from_file(inline: bool, cid: Optional[None]) -> None:
    ea = EmailAttachment.from_file(
        DATA_DIR / "sample.eml", inline=inline, content_id=cid
    )
    assert attr.asdict(ea, filter=lambda attr, _: attr.name != "content") == {
        "filename": "sample.eml",
        "inline": inline,
        "content_id": cid,
    }
    assert email2dict(ea.content) == {
        "unixfrom": None,
        "headers": {
            "from": [
                {
                    "display_name": "Steven E'Nder",
                    "address": "sender@example.nil",
                },
            ],
            "to": [
                {
                    "display_name": "",
                    "address": "recipient@redacted.nil",
                },
            ],
            "subject": "Seeking a job",
            "content-type": {
                "content_type": "multipart/mixed",
                "params": {},
            },
            "user-agent": [
                "Mutt/1.5.24 (2015-08-30)",
            ],
        },
        "preamble": "",
        "content": [
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "text/plain",
                        "params": {},
                    },
                    "content-disposition": {
                        "disposition": "inline",
                        "params": {},
                    },
                },
                "preamble": None,
                "content": "Please find my credentials attached.\n\n",
                "epilogue": None,
            },
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "text/plain",
                        "params": {"name": "résumé.txt"},
                    },
                    "content-disposition": {
                        "disposition": "attachment",
                        "params": {"filename": "résumé.txt"},
                    },
                },
                "preamble": None,
                "content": "- Wrote email2dict\n- Has a pulse (sometimes)\n",
                "epilogue": None,
            },
        ],
        "epilogue": "",
    }



from mailbits import email2dict
from eletter import Address, BytesAttachment, EmailAttachment, TextAttachment, compose


def test_compose_text_bytes_attachment_custom_content_type() -> None:
    msg = compose(
        from_="me@here.com",
        to=["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
        subject="Some electronic mail",
        text="This is the text of an e-mail.",
        attachments=[
            BytesAttachment(
                b"\xFE\xED\xFA\xCE",
                filename="blob.dat",
                content_type="application/x-feedface",
            )
        ],
    )
    assert email2dict(msg) == {
        "unixfrom": None,
        "headers": {
            "subject": "Some electronic mail",
            "from": [
                {
                    "display_name": "",
                    "address": "me@here.com",
                }
            ],
            "to": [
                {
                    "display_name": "",
                    "address": "you@there.net",
                },
                {
                    "display_name": "Thaddeus Hem",
                    "address": "them@hither.yon",
                },
            ],
            "content-type": {
                "content_type": "multipart/mixed",
                "params": {},
            },
        },
        "preamble": None,
        "content": [
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "text/plain",
                        "params": {},
                    },
                },
                "preamble": None,
                "content": "This is the text of an e-mail.\n",
                "epilogue": None,
            },
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "application/x-feedface",
                        "params": {},
                    },
                    "content-disposition": {
                        "disposition": "attachment",
                        "params": {"filename": "blob.dat"},
                    },
                },
                "preamble": None,
                "content": b"\xFE\xED\xFA\xCE",
                "epilogue": None,
            },
        ],
        "epilogue": None,
    }


def test_compose_text_bytes_attachment_set_custom_content_type() -> None:
    attach = BytesAttachment(b"\xFE\xED\xFA\xCE", filename="blob.dat")
    attach.content_type = "application/x-feedface; face=pretty"
    msg = compose(
        from_="me@here.com",
        to=["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
        subject="Some electronic mail",
        text="This is the text of an e-mail.",
        attachments=[attach],
    )
    assert email2dict(msg) == {
        "unixfrom": None,
        "headers": {
            "subject": "Some electronic mail",
            "from": [
                {
                    "display_name": "",
                    "address": "me@here.com",
                }
            ],
            "to": [
                {
                    "display_name": "",
                    "address": "you@there.net",
                },
                {
                    "display_name": "Thaddeus Hem",
                    "address": "them@hither.yon",
                },
            ],
            "content-type": {
                "content_type": "multipart/mixed",
                "params": {},
            },
        },
        "preamble": None,
        "content": [
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "text/plain",
                        "params": {},
                    },
                },
                "preamble": None,
                "content": "This is the text of an e-mail.\n",
                "epilogue": None,
            },
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "application/x-feedface",
                        "params": {"face": "pretty"},
                    },
                    "content-disposition": {
                        "disposition": "attachment",
                        "params": {"filename": "blob.dat"},
                    },
                },
                "preamble": None,
                "content": b"\xFE\xED\xFA\xCE",
                "epilogue": None,
            },
        ],
        "epilogue": None,
    }


def test_compose_text_bytes_attachment_custom_main_type() -> None:
    msg = compose(
        from_="me@here.com",
        to=["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
        subject="Some electronic mail",
        text="This is the text of an e-mail.",
        attachments=[
            BytesAttachment(
                b"\xFE\xED\xFA\xCE",
                filename="blob.dat",
                content_type="image/x-feedface",
            )
        ],
    )
    assert email2dict(msg) == {
        "unixfrom": None,
        "headers": {
            "subject": "Some electronic mail",
            "from": [
                {
                    "display_name": "",
                    "address": "me@here.com",
                }
            ],
            "to": [
                {
                    "display_name": "",
                    "address": "you@there.net",
                },
                {
                    "display_name": "Thaddeus Hem",
                    "address": "them@hither.yon",
                },
            ],
            "content-type": {
                "content_type": "multipart/mixed",
                "params": {},
            },
        },
        "preamble": None,
        "content": [
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "text/plain",
                        "params": {},
                    },
                },
                "preamble": None,
                "content": "This is the text of an e-mail.\n",
                "epilogue": None,
            },
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "image/x-feedface",
                        "params": {},
                    },
                    "content-disposition": {
                        "disposition": "attachment",
                        "params": {"filename": "blob.dat"},
                    },
                },
                "preamble": None,
                "content": b"\xFE\xED\xFA\xCE",
                "epilogue": None,
            },
        ],
        "epilogue": None,
    }


def test_compose_text_html_bytes_attachment() -> None:
    msg = compose(
        from_="me@here.com",
        to=["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
        subject="Some electronic mail",
        text="This is the text of an e-mail.",
        html="<p>This is the <i>text</i> of an <b>e</b>-mail.<p>",
        attachments=[
            BytesAttachment(
                b"\xFE\xED\xFA\xCE",
                filename="blob.dat",
                content_type="application/x-feedface",
            )
        ],
    )
    assert email2dict(msg) == {
        "unixfrom": None,
        "headers": {
            "subject": "Some electronic mail",
            "from": [
                {
                    "display_name": "",
                    "address": "me@here.com",
                }
            ],
            "to": [
                {
                    "display_name": "",
                    "address": "you@there.net",
                },
                {
                    "display_name": "Thaddeus Hem",
                    "address": "them@hither.yon",
                },
            ],
            "content-type": {
                "content_type": "multipart/mixed",
                "params": {},
            },
        },
        "preamble": None,
        "content": [
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "multipart/alternative",
                        "params": {},
                    },
                },
                "preamble": None,
                "content": [
                    {
                        "unixfrom": None,
                        "headers": {
                            "content-type": {
                                "content_type": "text/plain",
                                "params": {},
                            },
                        },
                        "preamble": None,
                        "content": "This is the text of an e-mail.\n",
                        "epilogue": None,
                    },
                    {
                        "unixfrom": None,
                        "headers": {
                            "content-type": {
                                "content_type": "text/html",
                                "params": {},
                            },
                        },
                        "preamble": None,
                        "content": (
                            "<p>This is the <i>text</i> of an <b>e</b>-mail.<p>\n"
                        ),
                        "epilogue": None,
                    },
                ],
                "epilogue": None,
            },
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "application/x-feedface",
                        "params": {},
                    },
                    "content-disposition": {
                        "disposition": "attachment",
                        "params": {"filename": "blob.dat"},
                    },
                },
                "preamble": None,
                "content": b"\xFE\xED\xFA\xCE",
                "epilogue": None,
            },
        ],
        "epilogue": None,
    }


def test_compose_text_bytes_attachment_default_content_type() -> None:
    msg = compose(
        from_="me@here.com",
        to=["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
        subject="Some electronic mail",
        text="This is the text of an e-mail.",
        attachments=[
            BytesAttachment(
                b"\xFE\xED\xFA\xCE",
                filename="blob.dat",
            )
        ],
    )
    assert email2dict(msg) == {
        "unixfrom": None,
        "headers": {
            "subject": "Some electronic mail",
            "from": [
                {
                    "display_name": "",
                    "address": "me@here.com",
                }
            ],
            "to": [
                {
                    "display_name": "",
                    "address": "you@there.net",
                },
                {
                    "display_name": "Thaddeus Hem",
                    "address": "them@hither.yon",
                },
            ],
            "content-type": {
                "content_type": "multipart/mixed",
                "params": {},
            },
        },
        "preamble": None,
        "content": [
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "text/plain",
                        "params": {},
                    },
                },
                "preamble": None,
                "content": "This is the text of an e-mail.\n",
                "epilogue": None,
            },
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "application/octet-stream",
                        "params": {},
                    },
                    "content-disposition": {
                        "disposition": "attachment",
                        "params": {"filename": "blob.dat"},
                    },
                },
                "preamble": None,
                "content": b"\xFE\xED\xFA\xCE",
                "epilogue": None,
            },
        ],
        "epilogue": None,
    }


def test_compose_text_bytes_attachment_paramed_content_type() -> None:
    msg = compose(
        from_="me@here.com",
        to=["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
        subject="Some electronic mail",
        text="This is the text of an e-mail.",
        attachments=[
            BytesAttachment(
                b"\xFE\xED\xFA\xCE",
                filename="blob.dat",
                content_type="application/x-feedface; face=pretty",
            )
        ],
    )
    assert email2dict(msg) == {
        "unixfrom": None,
        "headers": {
            "subject": "Some electronic mail",
            "from": [
                {
                    "display_name": "",
                    "address": "me@here.com",
                }
            ],
            "to": [
                {
                    "display_name": "",
                    "address": "you@there.net",
                },
                {
                    "display_name": "Thaddeus Hem",
                    "address": "them@hither.yon",
                },
            ],
            "content-type": {
                "content_type": "multipart/mixed",
                "params": {},
            },
        },
        "preamble": None,
        "content": [
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "text/plain",
                        "params": {},
                    },
                },
                "preamble": None,
                "content": "This is the text of an e-mail.\n",
                "epilogue": None,
            },
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "application/x-feedface",
                        "params": {"face": "pretty"},
                    },
                    "content-disposition": {
                        "disposition": "attachment",
                        "params": {"filename": "blob.dat"},
                    },
                },
                "preamble": None,
                "content": b"\xFE\xED\xFA\xCE",
                "epilogue": None,
            },
        ],
        "epilogue": None,
    }


def test_compose_text_bytes_attachment_inline() -> None:
    msg = compose(
        from_="me@here.com",
        to=["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
        subject="Some electronic mail",
        text="This is the text of an e-mail.",
        attachments=[
            BytesAttachment(
                b"\xFE\xED\xFA\xCE",
                filename="blob.dat",
                content_type="application/x-feedface",
                inline=True,
            )
        ],
    )
    assert email2dict(msg) == {
        "unixfrom": None,
        "headers": {
            "subject": "Some electronic mail",
            "from": [
                {
                    "display_name": "",
                    "address": "me@here.com",
                }
            ],
            "to": [
                {
                    "display_name": "",
                    "address": "you@there.net",
                },
                {
                    "display_name": "Thaddeus Hem",
                    "address": "them@hither.yon",
                },
            ],
            "content-type": {
                "content_type": "multipart/mixed",
                "params": {},
            },
        },
        "preamble": None,
        "content": [
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "text/plain",
                        "params": {},
                    },
                },
                "preamble": None,
                "content": "This is the text of an e-mail.\n",
                "epilogue": None,
            },
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "application/x-feedface",
                        "params": {},
                    },
                    "content-disposition": {
                        "disposition": "inline",
                        "params": {"filename": "blob.dat"},
                    },
                },
                "preamble": None,
                "content": b"\xFE\xED\xFA\xCE",
                "epilogue": None,
            },
        ],
        "epilogue": None,
    }


def test_compose_text_text_attachment_default_content_type() -> None:
    msg = compose(
        from_="me@here.com",
        to=["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
        subject="Some electronic mail",
        text="This is the text of an e-mail.",
        attachments=[TextAttachment("This is a text document.", filename="blob.txt")],
    )
    assert email2dict(msg) == {
        "unixfrom": None,
        "headers": {
            "subject": "Some electronic mail",
            "from": [
                {
                    "display_name": "",
                    "address": "me@here.com",
                }
            ],
            "to": [
                {
                    "display_name": "",
                    "address": "you@there.net",
                },
                {
                    "display_name": "Thaddeus Hem",
                    "address": "them@hither.yon",
                },
            ],
            "content-type": {
                "content_type": "multipart/mixed",
                "params": {},
            },
        },
        "preamble": None,
        "content": [
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "text/plain",
                        "params": {},
                    },
                },
                "preamble": None,
                "content": "This is the text of an e-mail.\n",
                "epilogue": None,
            },
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "text/plain",
                        "params": {},
                    },
                    "content-disposition": {
                        "disposition": "attachment",
                        "params": {"filename": "blob.txt"},
                    },
                },
                "preamble": None,
                "content": "This is a text document.\n",
                "epilogue": None,
            },
        ],
        "epilogue": None,
    }


def test_compose_text_text_attachment_custom_content_type() -> None:
    msg = compose(
        from_="me@here.com",
        to=["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
        subject="Some electronic mail",
        text="This is the text of an e-mail.",
        attachments=[
            TextAttachment(
                "This is a text document.",
                filename="blob.md",
                content_type="text/markdown",
            )
        ],
    )
    assert email2dict(msg) == {
        "unixfrom": None,
        "headers": {
            "subject": "Some electronic mail",
            "from": [
                {
                    "display_name": "",
                    "address": "me@here.com",
                }
            ],
            "to": [
                {
                    "display_name": "",
                    "address": "you@there.net",
                },
                {
                    "display_name": "Thaddeus Hem",
                    "address": "them@hither.yon",
                },
            ],
            "content-type": {
                "content_type": "multipart/mixed",
                "params": {},
            },
        },
        "preamble": None,
        "content": [
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "text/plain",
                        "params": {},
                    },
                },
                "preamble": None,
                "content": "This is the text of an e-mail.\n",
                "epilogue": None,
            },
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "text/markdown",
                        "params": {},
                    },
                    "content-disposition": {
                        "disposition": "attachment",
                        "params": {"filename": "blob.md"},
                    },
                },
                "preamble": None,
                "content": "This is a text document.\n",
                "epilogue": None,
            },
        ],
        "epilogue": None,
    }


def test_compose_text_text_attachment_set_custom_content_type() -> None:
    attach = TextAttachment("This is a text document.", filename="blob.md")
    attach.content_type = "text/markdown; variant=CommonMark"
    msg = compose(
        from_="me@here.com",
        to=["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
        subject="Some electronic mail",
        text="This is the text of an e-mail.",
        attachments=[attach],
    )
    assert email2dict(msg) == {
        "unixfrom": None,
        "headers": {
            "subject": "Some electronic mail",
            "from": [
                {
                    "display_name": "",
                    "address": "me@here.com",
                }
            ],
            "to": [
                {
                    "display_name": "",
                    "address": "you@there.net",
                },
                {
                    "display_name": "Thaddeus Hem",
                    "address": "them@hither.yon",
                },
            ],
            "content-type": {
                "content_type": "multipart/mixed",
                "params": {},
            },
        },
        "preamble": None,
        "content": [
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "text/plain",
                        "params": {},
                    },
                },
                "preamble": None,
                "content": "This is the text of an e-mail.\n",
                "epilogue": None,
            },
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "text/markdown",
                        "params": {"variant": "CommonMark"},
                    },
                    "content-disposition": {
                        "disposition": "attachment",
                        "params": {"filename": "blob.md"},
                    },
                },
                "preamble": None,
                "content": "This is a text document.\n",
                "epilogue": None,
            },
        ],
        "epilogue": None,
    }


def test_compose_text_text_attachment_paramed_content_type() -> None:
    msg = compose(
        from_="me@here.com",
        to=["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
        subject="Some electronic mail",
        text="This is the text of an e-mail.",
        attachments=[
            TextAttachment(
                "This is a text document.",
                filename="blob.md",
                content_type="text/markdown; variant=CommonMark",
            )
        ],
    )
    assert email2dict(msg) == {
        "unixfrom": None,
        "headers": {
            "subject": "Some electronic mail",
            "from": [
                {
                    "display_name": "",
                    "address": "me@here.com",
                }
            ],
            "to": [
                {
                    "display_name": "",
                    "address": "you@there.net",
                },
                {
                    "display_name": "Thaddeus Hem",
                    "address": "them@hither.yon",
                },
            ],
            "content-type": {
                "content_type": "multipart/mixed",
                "params": {},
            },
        },
        "preamble": None,
        "content": [
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "text/plain",
                        "params": {},
                    },
                },
                "preamble": None,
                "content": "This is the text of an e-mail.\n",
                "epilogue": None,
            },
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "text/markdown",
                        "params": {"variant": "CommonMark"},
                    },
                    "content-disposition": {
                        "disposition": "attachment",
                        "params": {"filename": "blob.md"},
                    },
                },
                "preamble": None,
                "content": "This is a text document.\n",
                "epilogue": None,
            },
        ],
        "epilogue": None,
    }


def test_compose_text_text_attachment_inline() -> None:
    msg = compose(
        from_="me@here.com",
        to=["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
        subject="Some electronic mail",
        text="This is the text of an e-mail.",
        attachments=[
            TextAttachment("This is a text document.", filename="blob.txt", inline=True)
        ],
    )
    assert email2dict(msg) == {
        "unixfrom": None,
        "headers": {
            "subject": "Some electronic mail",
            "from": [
                {
                    "display_name": "",
                    "address": "me@here.com",
                }
            ],
            "to": [
                {
                    "display_name": "",
                    "address": "you@there.net",
                },
                {
                    "display_name": "Thaddeus Hem",
                    "address": "them@hither.yon",
                },
            ],
            "content-type": {
                "content_type": "multipart/mixed",
                "params": {},
            },
        },
        "preamble": None,
        "content": [
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "text/plain",
                        "params": {},
                    },
                },
                "preamble": None,
                "content": "This is the text of an e-mail.\n",
                "epilogue": None,
            },
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "text/plain",
                        "params": {},
                    },
                    "content-disposition": {
                        "disposition": "inline",
                        "params": {"filename": "blob.txt"},
                    },
                },
                "preamble": None,
                "content": "This is a text document.\n",
                "epilogue": None,
            },
        ],
        "epilogue": None,
    }


def test_compose_text_text_attachment_charset() -> None:
    msg = compose(
        from_="me@here.com",
        to=["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
        subject="Some electronic mail",
        text="This is the text of an e-mail.",
        attachments=[
            TextAttachment(
                "This is a text document.",
                filename="blob.txt",
                content_type="text/plain; charset=iso-8859-1",
            )
        ],
    )
    assert email2dict(msg) == {
        "unixfrom": None,
        "headers": {
            "subject": "Some electronic mail",
            "from": [
                {
                    "display_name": "",
                    "address": "me@here.com",
                }
            ],
            "to": [
                {
                    "display_name": "",
                    "address": "you@there.net",
                },
                {
                    "display_name": "Thaddeus Hem",
                    "address": "them@hither.yon",
                },
            ],
            "content-type": {
                "content_type": "multipart/mixed",
                "params": {},
            },
        },
        "preamble": None,
        "content": [
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "text/plain",
                        "params": {},
                    },
                },
                "preamble": None,
                "content": "This is the text of an e-mail.\n",
                "epilogue": None,
            },
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "text/plain",
                        "params": {},
                    },
                    "content-disposition": {
                        "disposition": "attachment",
                        "params": {"filename": "blob.txt"},
                    },
                },
                "preamble": None,
                "content": "This is a text document.\n",
                "epilogue": None,
            },
        ],
        "epilogue": None,
    }
    assert email2dict(msg, include_all=True)["content"][1]["headers"]["content-type"][
        "params"
    ] == {"charset": "iso-8859-1"}


def test_compose_html_bytes_attachment_text_attachment() -> None:
    msg = compose(
        from_="me@here.com",
        to=["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
        subject="Some electronic mail",
        html="<p>This is the <i>text</i> of an <b>e</b>-mail.<p>",
        attachments=[
            BytesAttachment(
                b"\xFE\xED\xFA\xCE",
                filename="blob.dat",
                content_type="application/x-feedface",
            ),
            TextAttachment("This is a text document.", filename="foo.txt"),
        ],
    )
    assert email2dict(msg) == {
        "unixfrom": None,
        "headers": {
            "subject": "Some electronic mail",
            "from": [
                {
                    "display_name": "",
                    "address": "me@here.com",
                }
            ],
            "to": [
                {
                    "display_name": "",
                    "address": "you@there.net",
                },
                {
                    "display_name": "Thaddeus Hem",
                    "address": "them@hither.yon",
                },
            ],
            "content-type": {
                "content_type": "multipart/mixed",
                "params": {},
            },
        },
        "preamble": None,
        "content": [
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "text/html",
                        "params": {},
                    },
                },
                "preamble": None,
                "content": "<p>This is the <i>text</i> of an <b>e</b>-mail.<p>\n",
                "epilogue": None,
            },
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "application/x-feedface",
                        "params": {},
                    },
                    "content-disposition": {
                        "disposition": "attachment",
                        "params": {"filename": "blob.dat"},
                    },
                },
                "preamble": None,
                "content": b"\xFE\xED\xFA\xCE",
                "epilogue": None,
            },
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "text/plain",
                        "params": {},
                    },
                    "content-disposition": {
                        "disposition": "attachment",
                        "params": {"filename": "foo.txt"},
                    },
                },
                "preamble": None,
                "content": "This is a text document.\n",
                "epilogue": None,
            },
        ],
        "epilogue": None,
    }


def test_compose_email_attachment() -> None:
    msg = compose(
        from_="me@here.com",
        to=["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
        subject="Some electronic mail",
        text="Here's that e-mail you wanted me to send.",
        attachments=[
            EmailAttachment(
                compose(
                    from_="somebody@some.where",
                    to=["me@here.com"],
                    subject="A pretty picture!",
                    text="Look at the pretty picture!  ... It came through, right?",
                    attachments=[
                        BytesAttachment(
                            b"\xFE\xED\xFA\xCE",
                            filename="blob.dat",
                            content_type="application/x-feedface",
                        )
                    ],
                ),
                filename="message.eml",
                inline=False,
            )
        ],
    )
    assert email2dict(msg) == {
        "unixfrom": None,
        "headers": {
            "subject": "Some electronic mail",
            "from": [
                {
                    "display_name": "",
                    "address": "me@here.com",
                }
            ],
            "to": [
                {
                    "display_name": "",
                    "address": "you@there.net",
                },
                {
                    "display_name": "Thaddeus Hem",
                    "address": "them@hither.yon",
                },
            ],
            "content-type": {
                "content_type": "multipart/mixed",
                "params": {},
            },
        },
        "preamble": None,
        "content": [
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "text/plain",
                        "params": {},
                    },
                },
                "preamble": None,
                "content": "Here's that e-mail you wanted me to send.\n",
                "epilogue": None,
            },
            {
                "unixfrom": None,
                "headers": {
                    "content-type": {
                        "content_type": "message/rfc822",
                        "params": {},
                    },
                    "content-disposition": {
                        "disposition": "attachment",
                        "params": {"filename": "message.eml"},
                    },
                },
                "preamble": None,
                "content": {
                    "unixfrom": None,
                    "headers": {
                        "subject": "A pretty picture!",
                        "from": [
                            {
                                "display_name": "",
                                "address": "somebody@some.where",
                            },
                        ],
                        "to": [
                            {
                                "display_name": "",
                                "address": "me@here.com",
                            },
                        ],
                        "content-type": {
                            "content_type": "multipart/mixed",
                            "params": {},
                        },
                    },
                    "preamble": None,
                    "content": [
                        {
                            "unixfrom": None,
                            "headers": {
                                "content-type": {
                                    "content_type": "text/plain",
                                    "params": {},
                                },
                            },
                            "preamble": None,
                            "content": (
                                "Look at the pretty picture!  ... It came"
                                " through, right?\n"
                            ),
                            "epilogue": None,
                        },
                        {
                            "unixfrom": None,
                            "headers": {
                                "content-type": {
                                    "content_type": "application/x-feedface",
                                    "params": {},
                                },
                                "content-disposition": {
                                    "disposition": "attachment",
                                    "params": {"filename": "blob.dat"},
                                },
                            },
                            "preamble": None,
                            "content": b"\xFE\xED\xFA\xCE",
                            "epilogue": None,
                        },
                    ],
                    "epilogue": None,
                },
                "epilogue": None,
            },
        ],
        "epilogue": None,
    }



from datetime import datetime
from email.message import EmailMessage
from typing import Iterable, Mapping, Optional, Union
from mailbits import ContentType
from .classes import Attachment, HTMLBody, MailItem, TextBody
from .util import AddressOrGroup, SingleAddress


def compose(
    *,
    subject: str,
    to: Iterable[AddressOrGroup],
    from_: Optional[Union[AddressOrGroup, Iterable[AddressOrGroup]]] = None,
    text: Optional[str] = None,
    html: Optional[str] = None,
    cc: Optional[Iterable[AddressOrGroup]] = None,
    bcc: Optional[Iterable[AddressOrGroup]] = None,
    reply_to: Optional[Union[AddressOrGroup, Iterable[AddressOrGroup]]] = None,
    sender: Optional[SingleAddress] = None,
    date: Optional[datetime] = None,
    headers: Optional[Mapping[str, Union[str, Iterable[str]]]] = None,
    attachments: Optional[Iterable[Attachment]] = None,
) -> EmailMessage:
    """
    Construct an `~email.message.EmailMessage` instance from a subject,
    :mailheader:`From` address, :mailheader:`To` addresses, and a plain text
    and/or HTML body, optionally accompanied by attachments and other headers.

    All parameters other than ``subject``, ``to``, and at least one of ``text``
    and ``html`` are optional.

    :param str subject: The e-mail's :mailheader:`Subject` line
    :param to: The e-mail's :mailheader:`To` line
    :type to: iterable of addresses
    :param from_:
        The e-mail's :mailheader:`From` line.  Note that this argument is
        spelled with an underscore, as "``from``" is a keyword in Python.
    :type from_: address or iterable of addresses
    :param str text:
        The contents of a :mimetype:`text/plain` body for the e-mail.  At least
        one of ``text`` and ``html`` must be specified.
    :param str html:
        The contents of a :mimetype:`text/html` body for the e-mail.  At least
        one of ``text`` and ``html`` must be specified.
    :param cc: The e-mail's :mailheader:`CC` line
    :type cc: iterable of addresses
    :param bcc: The e-mail's :mailheader:`BCC` line
    :type bcc: iterable of addresses
    :param reply_to: The e-mail's :mailheader:`Reply-To` line
    :type reply_to: address or iterable of addresses
    :param address sender:
        The e-mail's :mailheader:`Sender` line.  The address must be a string
        or `Address`, not a `Group`.
    :param datetime date: The e-mail's :mailheader:`Date` line
    :param attachments:
        A collection of :ref:`attachments <attachments>` to append to the
        e-mail
    :type attachments: iterable of attachments
    :param mapping headers:
        A collection of additional headers to add to the e-mail.  A header
        value may be either a single string or an iterable of strings to add
        multiple headers with the same name.  If you wish to set an
        otherwise-unsupported address header like :mailheader:`Resent-From` to
        a list of addresses, use the `format_addresses()` function to first
        convert the addresses to a string.
    :rtype: email.message.EmailMessage
    :raises ValueError: if neither ``text`` nor ``html`` is set
    """
    msg: Optional[MailItem] = None
    if text is not None:
        msg = TextBody(text)
    if html is not None:
        if msg is None:
            msg = HTMLBody(html)
        else:
            msg |= HTMLBody(html)
    if msg is None:
        raise ValueError("At least one of text and html must be non-None")
    if attachments is not None:
        for a in attachments:
            msg &= a
    return msg.compose(
        subject=subject,
        from_=from_,
        to=to,
        cc=cc,
        bcc=bcc,
        reply_to=reply_to,
        sender=sender,
        date=date,
        headers=headers,
    )


def assemble_content_type(maintype: str, subtype: str, **params: str) -> str:
    """
    Construct a :mailheader:`Content-Type` string from a maintype, subtype, and
    some number of parameters

    :raises ValueError: if ``f"{maintype}/{subtype}"`` is an invalid
        :mailheader:`Content-Type`
    """
    return str(ContentType(maintype, subtype, params))


def reply_quote(s: str, prefix: str = "> ") -> str:
    """
    Quote__ a text following the *de facto* standard for replying to an e-mail;
    that is, prefix each line of the text with ``"> "`` (or a custom prefix),
    and if a line already starts with the prefix, omit any trailing whitespace
    from the newly-added prefix (so ``"> already quoted"`` becomes ``">>
    already quoted"``).

    If the resulting string does not end with a newline, one is added.  The
    empty string is treated as a single line.

    __ https://en.wikipedia.org/wiki/Usenet_quoting
    """
    s2 = ""
    for ln in (s or "\n").splitlines(True):
        if ln.startswith(prefix):
            s2 += prefix.rstrip() + ln
        else:
            s2 += prefix + ln
    if not s2.endswith(("\n", "\r")):
        s2 += "\n"
    return s2



"""
Simple e-mail composition

``eletter`` provides a basic function for constructing an
`email.message.EmailMessage` instance without having to touch the needlessly
complicated `EmailMessage` class itself.  E-mails with text bodies and/or HTML
bodies plus attachments are supported.  Classes are also provided for composing
more complex multipart e-mails.

Visit <https://github.com/jwodder/eletter> or <https://eletter.rtfd.io> for
more information.
"""

__version__ = "0.4.0"
__author__ = "John Thorvald Wodder II"
__author_email__ = "eletter@varonathe.org"
__license__ = "MIT"
__url__ = "https://github.com/jwodder/eletter"

from mailbits import format_addresses
from .classes import (
    Address,
    Alternative,
    Attachment,
    BytesAttachment,
    EmailAttachment,
    Group,
    HTMLBody,
    MailItem,
    Mixed,
    Multipart,
    Related,
    TextAttachment,
    TextBody,
)
from .core import assemble_content_type, compose, reply_quote

__all__ = [
    "Address",
    "Alternative",
    "Attachment",
    "BytesAttachment",
    "EmailAttachment",
    "Group",
    "HTMLBody",
    "MailItem",
    "Mixed",
    "Multipart",
    "Related",
    "TextAttachment",
    "TextBody",
    "assemble_content_type",
    "compose",
    "format_addresses",
    "reply_quote",
]
